# -*- coding: utf-8 -*-
from __future__ import print_function
import logging

import numpy as np

from keras.engine.training import Model
from keras.layers import Dense
from keras_wrapper.dataset import Dataset, saveDataset, loadDataset

logging.basicConfig(level=logging.INFO, format='[%(asctime)s] %(message)s', datefmt='%d/%m/%Y %H:%M:%S')
logger = logging.getLogger(__name__)


def test_models_allclose(model, model_init=None, model_next=None, rtol=1e-05, atol=1e-08, verbose=0):
    if isinstance(model, str):
        from keras_wrapper.cnn_model import loadModel
        model = loadModel(model, -1, full_path=True)
        model_init = model.model_init
        model_next = model.model_next
        model = model.model
    logger.info("Checking all models are close")
    model_names = map(str, model.weights)
    if model_init is None and model_next is None:
        logger.warning("Checking of models_allclose won't be performed, because model_init and model_next are None")
        return True

    if verbose > 0:
        print ("Checking model next weights")

    if model_next is not None:
        model_next_names = map(str, model_next.weights)
        for (index_next, name) in list(enumerate(model_next_names)):
            index_model = model_names.index(name)
            if not np.allclose(model.weights[index_model].get_value(), model_next.weights[index_next].get_value(), rtol=rtol, atol=atol):
                raise AssertionError('Parameters ' + name + ' are not close! (model index: ' + str(index_model) + ' model_next index ' + str(index_next) + ')')
            if verbose > 0:
                print ("Weights", name, "(position ", index_next, "at model_next - position", index_model, "at model are close")
    if verbose > 0:
        print ("Checking model init weights")

    if model_init is not None:
        model_init_names = map(str, model_init.weights)
        for (index_init, name) in list(enumerate(model_init_names)):
            index_model = model_names.index(name)
            if not np.allclose(model.weights[index_model].get_value(), model_init.weights[index_init].get_value(), rtol=rtol, atol=atol):
                raise AssertionError('Parameters ' + name + ' are not close! (model index: ' + str(index_model) + ' model_init index ' + str(index_init) + ')')
            if verbose > 0:
                print ("Weights", name, "(position ", index_init, "at model_init - position", index_model, "at model are close")
    return True


def test_2models_allclose(model1, model2, rtol=1e-05, atol=1e-08, verbose=0):
    if isinstance(model1, str):
        from keras_wrapper.cnn_model import loadModel
        model1 = loadModel(model1, -1, full_path=True)
        model1_init = model1.model_init
        model1_next = model1.model_next
        model1 = model1.model
    if isinstance(model2, str):
        from keras_wrapper.cnn_model import loadModel
        model2 = loadModel(model2, -1, full_path=True)
        model2_init = model2.model_init
        model2_next = model2.model_next
        model2 = model2.model

    logger.info("Checking all models (from model 1) are close...")
    if test_models_allclose(model1, model1_init, model1_next, rtol=rtol, atol=atol, verbose=verbose):
        logger.info("All close")
    else:
        logger.info("Not close!")

    logger.info("Checking all models (from model 2) are close...")
    if test_models_allclose(model2, model2_init, model2_next, rtol=rtol, atol=atol, verbose=verbose):
        logger.info("All close")
    else:
        logger.info("Not close!")

    model1_names = map(str, model1.weights)
    model2_names = map(str, model2.weights)

    if verbose > 0:
        print ("===========================")
        print ("Checking model weights")

    logger.info("Checking model_next 1 is close to model_next 2")
    if model1 is not None:
        model_next_names = map(str, model1.weights)
        for (index_next, name) in list(enumerate(model_next_names)):
            index_model = model2_names.index(name)
            if not np.allclose(model2.weights[index_model].get_value(), model1.weights[index_next].get_value(), rtol=rtol, atol=atol):
                raise AssertionError('Parameters ' + name + ' are not close! (model2 index: ' + str(index_model) + ' model1 index ' + str(index_next) + ')')
            if verbose > 0:
                print ("Weights", name, "(position ", index_next, "at model_next - position", index_model, "at model are close")

    if model2 is not None:
        model_next_names = map(str, model2.weights)
        for (index_next, name) in list(enumerate(model_next_names)):
            index_model = model1_names.index(name)
            if not np.allclose(model1.weights[index_model].get_value(), model2.weights[index_next].get_value(), rtol=rtol, atol=atol):
                raise AssertionError('Parameters ' + name + ' are not close! (model1 index: ' + str(index_model) + ' model2 index ' + str(index_next) + ')')
            if verbose > 0:
                print ("Weights", name, "(position ", index_next, "at model_next - position", index_model, "at model are close")

    if verbose > 0:
        print ("Checking model next weights")

    logger.info("Checking model_next 1 is close to model_next 2")
    if model1_next is not None:
        model_next_names = map(str, model1_next.weights)
        for (index_next, name) in list(enumerate(model_next_names)):
            index_model = model2_names.index(name)
            if not np.allclose(model2.weights[index_model].get_value(), model1_next.weights[index_next].get_value(), rtol=rtol, atol=atol):
                raise ('Parameters ' + name + ' are not close! (model2 index: ' + str(index_model) + ' model1_next index ' + str(index_next) + ')')
            if verbose > 0:
                print ("Weights", name, "(position ", index_next, "at model_next - position", index_model, "at model are close")

    if model2_next is not None:
        model_next_names = map(str, model2_next.weights)
        for (index_next, name) in list(enumerate(model_next_names)):
            index_model = model1_names.index(name)
            if not np.allclose(model1.weights[index_model].get_value(), model2_next.weights[index_next].get_value(), rtol=rtol, atol=atol):
                raise AssertionError('Parameters ' + name + ' are not close! (model1 index: ' + str(index_model) + ' model2_next index ' + str(index_next) + ')')
            if verbose > 0:
                print ("Weights", name, "(position ", index_next, "at model_next - position", index_model, "at model are close")

    if verbose > 0:
        print ("Checking model init weights")

    logger.info("Checking model_next 1 is close to model_next 2")
    if model1_init is not None:
        model_next_names = map(str, model1_init.weights)
        for (index_next, name) in list(enumerate(model_next_names)):
            index_model = model2_names.index(name)
            if not np.allclose(model2.weights[index_model].get_value(), model1_init.weights[index_next].get_value(), rtol=rtol, atol=atol):
                raise AssertionError('Parameters ' + name + ' are not close! (model2 index: ' + str(index_model) + ' model1_init index ' + str(index_next) + ')')
            if verbose > 0:
                print ("Weights", name, "(position ", index_next, "at model_next - position", index_model, "at model are close")

    if model2_init is not None:
        model_next_names = map(str, model2_init.weights)
        for (index_next, name) in list(enumerate(model_next_names)):
            index_model = model1_names.index(name)
            if not np.allclose(model1.weights[index_model].get_value(), model2_init.weights[index_next].get_value(), rtol=rtol, atol=atol):
                raise AssertionError('Parameters ' + name + ' are not close! (model1 index: ' + str(index_model) + ' model2_next index ' + str(index_next) + ')')
            if verbose > 0:
                print ("Weights", name, "(position ", index_next, "at model_next - position", index_model, "at model are close")

    return True


def main_test():
    # loadFlickr8k() # load Flickr8k dataset for Image Description

    # loadMSVD() # load MSVD dataset for Video Description

    loadFood101()  # load Food101 dataset for Image Classification

    # Build basic model for image classification
    classifyFood101()


#################################
#
#    Model building functions
#
#################################

def classifyFood101():
    from keras_wrapper.cnn_model import CNN_Model, loadModel, saveModel
    logger.info('Defining CNN model and training it.')

    # Load food classification dataset
    dataset_name = 'Food101'
    ds = loadDataset('Datasets/Dataset_' + dataset_name + '.pkl')
    # The network we are going to use needs an image input size of [224,224,3]
    # for this reason we have to communicate this to the dataset instance in charge of loading the data
    ds.img_size_crop['images'] = [224, 224, 3]

    # Create VGG model and load weights
    model_name = 'VGG_16_FunctionalAPI'
    net = CNN_Model(type='VGG_16_FunctionalAPI', model_name=model_name, input_shape=[224, 224, 3],
                    weights_path='/media/HDD_2TB/CNN_MODELS/VGG/vgg16_weights.h5',
                    seq_to_functional=True)  # we are setting the weights of a Sequential model into a FunctionalAPI one

    # Reformat net output layer for the number of classes in our dataset
    n_classes = len(ds.classes['labels'])
    vis_input = net.model.get_layer('vis_input').output  # input layer
    drop = net.model.get_layer('last_dropout').output  # layer before final FC
    output = Dense(n_classes, activation='softmax', name='output')(drop)  # redefine FC-softmax layer
    net.model = Model(input=vis_input, output=output)  # define inputs and outputs

    # Compile
    net.setOptimizer(lr=0.001, metrics=['accuracy'])

    # Define the inputs and outputs mapping from our Dataset instance to our CNN_Model instance
    # set input and output mappings from dataset to network
    pos_images = ds.types_inputs.index('image')
    pos_labels = ds.types_outputs.index('categorical')

    # the first input of our dataset (pos_images) will also be the first input of our model (named vis_input)
    inputMapping = {'vis_input': pos_images}
    net.setInputsMapping(inputMapping)

    # the first output of our dataset (pos_labels) will also be the first output of our model (named output)
    outputMapping = {'output': pos_labels}
    net.setOutputsMapping(outputMapping, acc_output='output')

    # Save model
    saveModel(net, 0)

    # Load model
    net = loadModel('Models/' + model_name, 0)
    # the model must be compiled again when loaded
    net.setOptimizer(lr=0.001, metrics=['accuracy'])

    # Apply short training (1 epoch)
    # training_params = {'n_epochs': 1, 'batch_size': 50,
    #                    'lr_decay': 2, 'lr_gamma': 0.8,
    #                    'epochs_for_save': 1, 'verbose': 1, 'eval_on_sets': ['val']}
    # net.trainNet(ds, training_params)

    # Test network on test set
    test_params = {'batch_size': 50}
    # net.testNet(ds, test_params)

    # Predict network on all sets
    test_params['predict_on_sets'] = ['val']
    predictions = net.predictNet(ds, test_params)
    logger.info("Predicted %d samples." % (len(predictions)))
    logger.info("Done")


#################################
#
#    Data loading functions
#
#################################

def loadFlickr8k():
    logger.info('Loading Flickr8k dataset')

    # Build basic dataset structure
    #    we assign it a name and the path were the images are stored

    base_path = '/media/HDD_2TB/DATASETS/Flickr8k/'
    name = 'Flickr8k_ImageDescription'
    ds = Dataset(name, base_path + 'Flicker8k_Dataset')
    max_text_len = 35

    # Let's load the train, val and test splits of the descriptions (outputs)
    #    the files include a description per line
    #    and a set of 5 consecutive descriptions correspond to a single input image

    ds.setOutput(base_path + 'text/train_descriptions.txt', 'train',
                 type='text', id='descriptions',
                 tokenization='tokenize_basic', build_vocabulary=True, max_text_len=max_text_len)
    ds.setOutput(base_path + 'text/val_descriptions.txt', 'val',
                 type='text', id='descriptions',
                 tokenization='tokenize_basic', max_text_len=max_text_len)
    ds.setOutput(base_path + 'text/test_descriptions.txt', 'test',
                 type='text', id='descriptions',
                 tokenization='tokenize_basic', max_text_len=max_text_len)

    # Let's load the associated images (inputs)
    #    we must take into account that in this dataset we have 5 sentences per image,
    #    for this reason we introduce the parameter 'repeat_set'=5

    ds.setInput(base_path + 'text/Flickr_8k.trainImages.txt', 'train', type='image', id='images', repeat_set=5)
    ds.setInput(base_path + 'text/Flickr_8k.devImages.txt', 'val', type='image', id='images', repeat_set=5)
    ds.setInput(base_path + 'text/Flickr_8k.testImages.txt', 'test', type='image', id='images', repeat_set=5)

    # Now let's set the dataset mean image for preprocessing the data
    ds.setTrainMean(mean_image=[122.6795, 116.6690, 104.0067], data_id='images')

    # We have finished loading the dataset, now we can store it for using it in the future
    saveDataset(ds, 'Datasets')

    # We can easily recover it with a single line
    ds = loadDataset('Datasets/Dataset_' + name + '.pkl')

    # Lets recover the first batch of data
    [X, Y] = ds.getXY('train', 10)
    logger.info('Sample data loaded correctly. %d input samples. %d output samples' % (len(X), len(Y)))


def loadMSVD():
    logger.info('Loading MSVD dataset')

    # Build basic dataset structure
    #    we assign it a name and the path were the images are stored

    base_path = '/media/HDD_2TB/DATASETS/MSVD/'
    name = 'MSVD_VideoDescription'
    ds = Dataset(name, base_path)
    max_text_len = 35

    # Let's load the train, val and test splits of the descriptions (outputs)
    #    the files include a description per line. In this dataset a variable number
    #    of descriptions per video are provided.

    ds.setOutput(base_path + 'train_descriptions.txt', 'train',
                 type='text', id='descriptions',
                 tokenization='tokenize_basic', build_vocabulary=True, max_text_len=max_text_len)
    ds.setOutput(base_path + 'val_descriptions.txt', 'val',
                 type='text', id='descriptions',
                 tokenization='tokenize_basic', max_text_len=max_text_len)
    ds.setOutput(base_path + 'test_descriptions.txt', 'test',
                 type='text', id='descriptions',
                 tokenization='tokenize_basic', max_text_len=max_text_len)

    # Let's load the associated videos (inputs)
    #    we must take into account that in this dataset we have a different number of sentences per video,
    #    for this reason we introduce the parameter 'repeat_set'=num_captions, where num_captions is a list
    #    containing the number of captions in each video.

    num_captions_train = np.load(base_path + 'train_descriptions_counts.npy')
    num_captions_val = np.load(base_path + 'val_descriptions_counts.npy')
    num_captions_test = np.load(base_path + 'test_descriptions_counts.npy')

    ds.setInput([base_path + 'train_imgs_list.txt', base_path + 'train_imgs_counts.txt'],
                'train', type='video', id='videos',
                repeat_set=num_captions_train)
    ds.setInput([base_path + 'val_imgs_list.txt', base_path + 'val_imgs_counts.txt'],
                'val', type='video', id='videos',
                repeat_set=num_captions_val)
    ds.setInput([base_path + 'test_imgs_list.txt', base_path + 'test_imgs_counts.txt'],
                'test', type='video', id='videos',
                repeat_set=num_captions_test)

    # Now let's set the dataset mean image for preprocessing the data
    ds.setTrainMean(mean_image=[122.6795, 116.6690, 104.0067], data_id='videos')

    # We have finished loading the dataset, now we can store it for using it in the future
    saveDataset(ds, 'Datasets')

    # We can easily recover it with a single line
    ds = loadDataset('Datasets/Dataset_' + name + '.pkl')

    # Lets recover the first batch of data
    [X, Y] = ds.getXY('train', 10)
    logger.info('Sample data loaded correctly.')


def loadFood101():
    logger.info('Loading Food101 dataset')
    logger.info(
        'INFO: in order to load this dataset it must be placed in ../data/Food101/images/ after downloading it form https://www.vision.ee.ethz.ch/datasets_extra/food-101/')

    base_path = '../data/Food101/'
    name = 'Food101'
    ds = Dataset(name, base_path + 'images')

    # Insert inputs (images)
    ds.setInput(base_path + 'meta/train_split.txt', 'train',
                type='image', id='images', img_size_crop=[227, 227, 3])
    ds.setInput(base_path + 'meta/val_split.txt', 'val',
                type='image', id='images')
    ds.setInput(base_path + 'meta/test.txt', 'test',
                type='image', id='images')

    # Insert outputs (labels)
    ds.setOutput(base_path + 'meta/train_labels.txt', 'train',
                 type='categorical', id='labels')
    ds.setOutput(base_path + 'meta/val_labels.txt', 'val',
                 type='categorical', id='labels')
    ds.setOutput(base_path + 'meta/test_labels.txt', 'test',
                 type='categorical', id='labels')

    # Set list of classes (strings)
    ds.setClasses(base_path + 'meta/classes.txt', 'labels')

    # Now let's set the dataset mean image for preprocessing the data
    ds.setTrainMean(mean_image=[122.6795, 116.6690, 104.0067], data_id='images')

    # We have finished loading the dataset, now we can store it for using it in the future
    saveDataset(ds, 'Datasets')

    # We can easily recover it with a single line
    ds = loadDataset('Datasets/Dataset_' + name + '.pkl')

    # Lets recover the first batch of data
    [X, Y] = ds.getXY('train', 10)
    logger.info('Sample data loaded correctly.')



# -*- coding: utf-8 -*-
import re
import sys


def tokenize_basic(caption, lowercase=True):
    """
    Basic tokenizer for the input/output data of type 'text':
       * Splits punctuation
       * Optional lowercasing

    :param caption: String to tokenize
    :param lowercase: Whether to lowercase the caption or not
    :return: Tokenized version of caption
    """

    punct = [u'.',
             u';',
             u',',
             u'，',
             u"/",
             u'[',
             u']',
             u'"',
             u'{',
             u'}',
             u'(',
             u')',
             u'=',
             u'+',
             u'\\',
             u'_',
             u'-',
             u'>',
             u'<',
             u'@',
             u'`',
             u'¿',
             u'?',
             u'¡',
             u'!']

    def processPunctuation(inText):
        """
        Removes punctuation.
        """
        outText = inText
        for p in punct:
            outText = outText.replace(p, u' ' + p + u' ')
        return outText

    resAns = caption.lower() if lowercase else caption
    resAns = resAns.replace(u'\n', u' ')
    resAns = resAns.replace(u'\t', u' ')
    resAns = processPunctuation(resAns)
    resAns = re.sub(u'[ ]+', u' ', resAns)
    return resAns


def tokenize_aggressive(caption, lowercase=True):
    """
    Aggressive tokenizer for the input/output data of type 'text':
    * Removes punctuation
    * Optional lowercasing

    :param caption: String to tokenize
    :param lowercase: Whether to lowercase the caption or not
    :return: Tokenized version of caption
    """
    punct = [u'.',
             u';',
             u',',
             u'，',
             u"/",
             u'[',
             u']',
             u'"',
             u'{',
             u'}',
             u'(',
             u')',
             u'=',
             u'+',
             u'\\',
             u'_',
             u'-',
             u'>',
             u'<',
             u'@',
             u'`',
             u'¿',
             u'?',
             u'¡',
             u'!']

    def processPunctuation(inText):
        outText = inText
        for p in punct:
            outText = outText.replace(p, '')
        return outText

    resAns = caption.lower() if lowercase else caption
    resAns = processPunctuation(resAns)
    resAns = re.sub('[ ]+', ' ', resAns)
    resAns = resAns.strip()
    return resAns


def tokenize_icann(caption):
    """
    Tokenization used for the icann paper:
    * Removes some punctuation (. , ")
    * Lowercasing

    :param caption: String to tokenize
    :return: Tokenized version of caption
    """
    return tokenize_soft(caption)


def tokenize_montreal(caption):
    """
    Similar to tokenize_icann
        * Removes some punctuation
        * Lowercase

    :param caption: String to tokenize
    :return: Tokenized version of caption
    """
    tokenized = re.sub('[.,"\n\t]+', '', caption.strip())
    tokenized = re.sub('[\']+', " '", tokenized)
    tokenized = re.sub('[ ]+', ' ', tokenized)
    tokenized = map(lambda x: x.lower(), tokenized.split())
    tokenized = " ".join(tokenized)
    return tokenized


def tokenize_soft(caption, lowercase=True):
    """
    Tokenization used for the icann paper:
        * Removes very little punctuation
        * Lowercase
    :param caption: String to tokenize
    :param lowercase: Whether to lowercase the caption or not
    :return: Tokenized version of caption
    """
    tokenized = re.sub(u'[\n\t]+', u'', caption.strip())
    tokenized = re.sub(u'[\.]+', u' . ', tokenized)
    tokenized = re.sub(u'[,]+', u' , ', tokenized)
    tokenized = re.sub(u'[!]+', u' ! ', tokenized)
    tokenized = re.sub(u'[?]+', u' ? ', tokenized)
    tokenized = re.sub(u'[\{]+', u' { ', tokenized)
    tokenized = re.sub(u'[\}]+', u' } ', tokenized)
    tokenized = re.sub(u'[\(]+', u' ( ', tokenized)
    tokenized = re.sub(u'[\)]+', u' ) ', tokenized)
    tokenized = re.sub(u'[\[]+', u' [ ', tokenized)
    tokenized = re.sub(u'[\]]+', u' ] ', tokenized)
    tokenized = re.sub(u'["]+', u' " ', tokenized)
    tokenized = re.sub(u'[\']+', u" ' ", tokenized)
    tokenized = re.sub(u'[ ]+', u' ', tokenized)
    # tokenized = u" ".join(tokenized)
    return tokenized if not lowercase else tokenized.lower()


def tokenize_none(caption):
    """
    Does not tokenizes the sentences. Only performs a stripping

    :param caption: String to tokenize
    :return: Tokenized version of caption
    """
    tokenized = re.sub('[\n\t]+', '', caption.strip())
    return tokenized


def tokenize_none_char(caption):
    """
    Character-level tokenization. Respects all symbols. Separates chars. Inserts <space> sybmol for spaces.
    If found an escaped char, "&apos;" symbol, it is converted to the original one
    # List of escaped chars (by moses tokenizer)
    & ->  &amp;
    | ->  &#124;
    < ->  &lt;
    > ->  &gt;
    ' ->  &apos;
    " ->  &quot;
    [ ->  &#91;
    ] ->  &#93;
    :param caption: String to tokenize
    :return: Tokenized version of caption
    """

    def convert_chars(x):
        """
        Convert spaces to the '<space>' token.
        """
        if x == ' ':
            return u'<space>'
        else:
            return x

    tokenized = re.sub(u'[\n\t]+', u'', caption.strip())
    tokenized = re.sub(u'[ ]+', u' ', tokenized)
    tokenized = [convert_chars(char) for char in tokenized]
    tokenized = u" ".join(tokenized)
    return tokenized


def tokenize_CNN_sentence(caption):
    """
    Tokenization employed in the CNN_sentence package
    (https://github.com/yoonkim/CNN_sentence/blob/master/process_data.py#L97).
    :param caption: String to tokenize
    :return: Tokenized version of caption
    """
    tokenized = re.sub(r"[^A-Za-z0-9(),!?\'\`]", " ", caption)
    tokenized = re.sub(r"\'s", " \'s", tokenized)
    tokenized = re.sub(r"\'ve", " \'ve", tokenized)
    tokenized = re.sub(r"n\'t", " n\'t", tokenized)
    tokenized = re.sub(r"\'re", " \'re", tokenized)
    tokenized = re.sub(r"\'d", " \'d", tokenized)
    tokenized = re.sub(r"\'ll", " \'ll", tokenized)
    tokenized = re.sub(r",", " , ", tokenized)
    tokenized = re.sub(r"!", " ! ", tokenized)
    tokenized = re.sub(r"\(", " \( ", tokenized)
    tokenized = re.sub(r"\)", " \) ", tokenized)
    tokenized = re.sub(r"\?", " \? ", tokenized)
    tokenized = re.sub(r"\s{2,}", " ", tokenized)
    return tokenized.strip().lower()


def tokenize_questions(caption):
    """
    Basic tokenizer for VQA questions:
        * Lowercasing
        * Splits contractions
        * Removes punctuation
        * Numbers to digits

    :param caption: String to tokenize
    :return: Tokenized version of caption
    """
    contractions = {"aint": "ain't", "arent": "aren't", "cant": "can't", "couldve": "could've",
                    "couldnt": "couldn't",
                    "couldn'tve": "couldn’t’ve", "couldnt’ve": "couldn’t’ve", "didnt": "didn’t",
                    "doesnt": "doesn’t",
                    "dont": "don’t", "hadnt": "hadn’t", "hadnt’ve": "hadn’t’ve", "hadn'tve": "hadn’t’ve",
                    "hasnt": "hasn’t", "havent": "haven’t", "hed": "he’d", "hed’ve": "he’d’ve", "he’dve": "he’d’ve",
                    "hes": "he’s", "howd": "how’d", "howll": "how’ll", "hows": "how’s", "Id’ve": "I’d’ve",
                    "I’dve": "I’d’ve", "Im": "I’m", "Ive": "I’ve", "isnt": "isn’t", "itd": "it’d",
                    "itd’ve": "it’d’ve",
                    "it’dve": "it’d’ve", "itll": "it’ll", "let’s": "let’s", "maam": "ma’am", "mightnt": "mightn’t",
                    "mightnt’ve": "mightn’t’ve", "mightn’tve": "mightn’t’ve", "mightve": "might’ve",
                    "mustnt": "mustn’t",
                    "mustve": "must’ve", "neednt": "needn’t", "notve": "not’ve", "oclock": "o’clock",
                    "oughtnt": "oughtn’t",
                    "ow’s’at": "’ow’s’at", "’ows’at": "’ow’s’at", "’ow’sat": "’ow’s’at", "shant": "shan’t",
                    "shed’ve": "she’d’ve", "she’dve": "she’d’ve", "she’s": "she’s", "shouldve": "should’ve",
                    "shouldnt": "shouldn’t", "shouldnt’ve": "shouldn’t’ve", "shouldn’tve": "shouldn’t’ve",
                    "somebody’d": "somebodyd", "somebodyd’ve": "somebody’d’ve", "somebody’dve": "somebody’d’ve",
                    "somebodyll": "somebody’ll", "somebodys": "somebody’s", "someoned": "someone’d",
                    "someoned’ve": "someone’d’ve", "someone’dve": "someone’d’ve", "someonell": "someone’ll",
                    "someones": "someone’s", "somethingd": "something’d", "somethingd’ve": "something’d’ve",
                    "something’dve": "something’d’ve", "somethingll": "something’ll", "thats": "that’s",
                    "thered": "there’d", "thered’ve": "there’d’ve", "there’dve": "there’d’ve",
                    "therere": "there’re",
                    "theres": "there’s", "theyd": "they’d", "theyd’ve": "they’d’ve", "they’dve": "they’d’ve",
                    "theyll": "they’ll", "theyre": "they’re", "theyve": "they’ve", "twas": "’twas",
                    "wasnt": "wasn’t",
                    "wed’ve": "we’d’ve", "we’dve": "we’d’ve", "weve": "we've", "werent": "weren’t",
                    "whatll": "what’ll",
                    "whatre": "what’re", "whats": "what’s", "whatve": "what’ve", "whens": "when’s", "whered":
                        "where’d", "wheres": "where's", "whereve": "where’ve", "whod": "who’d",
                    "whod’ve": "who’d’ve",
                    "who’dve": "who’d’ve", "wholl": "who’ll", "whos": "who’s", "whove": "who've", "whyll": "why’ll",
                    "whyre": "why’re", "whys": "why’s", "wont": "won’t", "wouldve": "would’ve",
                    "wouldnt": "wouldn’t",
                    "wouldnt’ve": "wouldn’t’ve", "wouldn’tve": "wouldn’t’ve", "yall": "y’all",
                    "yall’ll": "y’all’ll",
                    "y’allll": "y’all’ll", "yall’d’ve": "y’all’d’ve", "y’alld’ve": "y’all’d’ve",
                    "y’all’dve": "y’all’d’ve",
                    "youd": "you’d", "youd’ve": "you’d’ve", "you’dve": "you’d’ve", "youll": "you’ll",
                    "youre": "you’re", "youve": "you’ve"}
    punct = [';', r"/", '[', ']', '"', '{', '}', '(', ')', '=', '+', '\\',
             '_', '-', '>', '<', '@', '`', ',', '?', '!']
    commaStrip = re.compile("(\d)(\,)(\d)")
    periodStrip = re.compile("(?!<=\d)(\.)(?!\d)")
    manualMap = {'none': '0', 'zero': '0', 'one': '1', 'two': '2', 'three': '3', 'four': '4', 'five': '5',
                 'six': '6', 'seven': '7', 'eight': '8', 'nine': '9', 'ten': '10'}
    articles = ['a', 'an', 'the']

    def processPunctuation(inText):
        """
        Process (remove) the punctuation.
        """
        if isinstance(inText, str) and sys.version_info < (3, 0):
            inText = inText.decode("utf-8").encode("utf-8")
        outText = inText
        for p in punct:
            if (p + ' ' in inText or ' ' + p in inText) or (re.search(commaStrip, inText) is not None):
                outText = outText.replace(p, '')
            else:
                outText = outText.replace(p, ' ')
        outText = periodStrip.sub("", outText, re.UNICODE)
        return outText

    def processDigitArticle(inText):
        outText = []
        tempText = inText.lower().split()
        for word in tempText:
            word = manualMap.setdefault(word, word)
            if word not in articles:
                outText.append(word)
            else:
                pass
        for wordId, word in list(enumerate(outText)):
            if word in contractions:
                outText[wordId] = contractions[word]
        outText = ' '.join(outText)
        return outText

    resAns = caption.lower()
    resAns = resAns.replace('\n', ' ')
    resAns = resAns.replace('\t', ' ')
    resAns = resAns.strip()
    resAns = processPunctuation(resAns)
    resAns = processDigitArticle(resAns)

    return resAns


def tokenize_bpe(self, caption):
    """
    Applies BPE segmentation (https://github.com/rsennrich/subword-nmt)
    :param caption: Caption to detokenize.
    :return: Encoded version of caption.
    """
    if not self.BPE_built:
        raise Exception('Prior to use the "tokenize_bpe" method, you should invoke "build_BPE"')
    if isinstance(caption, str) and sys.version_info < (3, 0):
        caption = caption.decode('utf-8')
    tokenized = re.sub(u'[\n\t]+', u'', caption)
    tokenized = self.BPE.segment(tokenized).strip()
    return tokenized


def detokenize_none(caption):
    """
    Dummy function: Keeps the caption as it is.
    :param caption: String to de-tokenize.
    :return: Same caption.
    """
    if isinstance(caption, str) and sys.version_info < (3, 0):
        caption = caption.decode('utf-8')
    return caption


def detokenize_bpe(caption, separator=u'@@'):
    """
    Reverts BPE segmentation (https://github.com/rsennrich/subword-nmt)
    :param caption: Caption to detokenize.
    :param separator: BPE separator.
    :return: Detokenized version of caption.
    """
    if isinstance(caption, str) and sys.version_info < (3, 0):
        caption = caption.decode('utf-8')
    bpe_detokenization = re.compile(u'(' + separator + u' )|(' + separator + u' ?$)')
    detokenized = bpe_detokenization.sub(u'', caption).strip()
    return detokenized


def detokenize_none_char(caption):
    """
    Character-level detokenization. Respects all symbols. Joins chars into words. Words are delimited by
    the <space> token. If found an special character is converted to the escaped char.
    # List of escaped chars (by moses tokenizer)
        & ->  &amp;
        | ->  &#124;
        < ->  &lt;
        > ->  &gt;
        ' ->  &apos;
        " ->  &quot;
        [ ->  &#91;
        ] ->  &#93;
    :param caption: String to de-tokenize.
        :return: Detokenized version of caption.
    """

    detokenized = re.sub(u' ', u'', caption)
    detokenized = re.sub(u'<space>', u' ', detokenized)
    return detokenized



# -*- coding: utf-8 -*-
import copy
import itertools
import sys
import time
from six import iteritems
import numpy as np
import logging

logging.basicConfig(level=logging.INFO, format='[%(asctime)s] %(message)s', datefmt='%d/%m/%Y %H:%M:%S')
logger = logging.getLogger(__name__)
if sys.version_info.major == 2:
    from itertools import imap as map


def checkParameters(input_params, default_params, hard_check=False):
    """Validates a set of input parameters and uses the default ones if not specified.

    :param input_params: Input parameters.
    :param default_params: Default parameters
    :param hard_check: If True, raise exception if a parameter is not valid.
    :return:
    """
    valid_params = [key for key in default_params]
    params = dict()

    # Check input parameters' validity
    for key, val in iteritems(input_params):
        if key in valid_params:
            params[key] = val
        elif hard_check:
            raise ValueError("Parameter '" + key + "' is not a valid parameter.")

    # Use default parameters if not provided
    for key, default_val in iteritems(default_params):
        if key not in params:
            params[key] = default_val

    return params


class MultiprocessQueue():
    """
        Wrapper class for encapsulating the behaviour of some multiprocessing
        communication structures.

        See how Queues and Pipes work in the following link
        https://docs.python.org/2/library/multiprocessing.html#multiprocessing-examples
    """

    def __init__(self, manager, multiprocess_type='Queue'):
        if multiprocess_type != 'Queue' and multiprocess_type != 'Pipe':
            raise NotImplementedError(
                'Not valid multiprocessing queue of type ' + multiprocess_type)

        self.type = multiprocess_type
        if multiprocess_type == 'Queue':
            self.queue = eval('manager.' + multiprocess_type + '()')
        else:
            self.queue = eval(multiprocess_type + '()')

    def put(self, elem):
        if self.type == 'Queue':
            self.queue.put(elem)
        elif self.type == 'Pipe':
            self.queue[1].send(elem)

    def get(self):
        if self.type == 'Queue':
            return self.queue.get()
        elif self.type == 'Pipe':
            return self.queue[0].recv()

    def qsize(self):
        if self.type == 'Queue':
            return self.queue.qsize()
        elif self.type == 'Pipe':
            return -1

    def empty(self):
        if self.type == 'Queue':
            return self.queue.empty()
        elif self.type == 'Pipe':
            return not self.queue[0].poll()


def bbox(img, mode='max'):
    """
    Returns a bounding box covering all the non-zero area in the image.

    :param img: Image on which print the bounding box
    :param mode:  "width_height" returns width in [2] and height in [3], "max" returns xmax in [2] and ymax in [3]
    :return:
    """
    rows = np.any(img, axis=1)
    cols = np.any(img, axis=0)
    y, ymax = np.where(rows)[0][[0, -1]]
    x, xmax = np.where(cols)[0][[0, -1]]

    if mode == 'width_height':
        return x, y, xmax - x, ymax - y
    elif mode == 'max':
        return x, y, xmax, ymax


def build_OneVsOneECOC_Stage(n_classes_ecoc, input_shape, ds, stage1_lr=0.01,
                             ecoc_version=2):
    """

    :param n_classes_ecoc:
    :param input_shape:
    :param ds:
    :param stage1_lr:
    :param ecoc_version:
    :return:
    """
    n_classes = len(ds.classes)
    labels_list = [str(l) for l in range(n_classes)]

    combs = tuple(itertools.combinations(labels_list, n_classes_ecoc))
    stage = list()
    outputs_list = list()

    count = 0
    n_combs = len(combs)
    for c in combs:
        t = time.time()

        # Create each one_vs_one classifier of the intermediate stage
        if ecoc_version == 1:
            s = Stage(nInput=n_classes, nOutput=n_classes_ecoc,
                      input_shape=input_shape, output_shape=[1, 2],
                      type='One_vs_One_Inception', silence=True)
        elif ecoc_version == 2:
            s = Stage(nInput=n_classes, nOutput=n_classes_ecoc,
                      input_shape=input_shape, output_shape=[1, 2],
                      type='One_vs_One_Inception_v2', silence=True)
        # Build input mapping
        input_mapping = dict()
        for i in range(n_classes):
            i_str = str(i)
            if i_str in c:
                input_mapping[i] = c.index(i_str)
            else:
                input_mapping[i] = None
        # Build output mask
        # output_mask = {'[0]': [0], '[1]': None}
        s.defineClassMapping(input_mapping)
        # s.defineOutputMask(output_mask)
        s.setOptimizer(lr=stage1_lr)
        s.silence = False
        stage.append(s)

        outputs_list.append('loss_OnevsOne/output')

        logger.info('Built model %s/%s for classes %s in %0.5s seconds.' % (str(count + 1), str(n_combs), c, str(time.time() - t)))
        count += 1

    return [stage, outputs_list]


def build_OneVsAllECOC_Stage(n_classes_ecoc, input_shape, ds, stage1_lr):
    """

    :param n_classes_ecoc:
    :param input_shape:
    :param ds:
    :param stage1_lr:
    :return:
    """
    n_classes = len(ds.classes)

    stage = list()
    outputs_list = list()

    count = 0
    for c in range(n_classes):
        t = time.time()

        # Create each one_vs_one classifier of the intermediate stage
        s = Stage(nInput=n_classes, nOutput=n_classes_ecoc, input_shape=input_shape,
                  output_shape=[1],
                  type='One_vs_One_Inception', silence=True)
        # Build input mapping
        input_mapping = dict()
        for i in range(n_classes):
            if i == c:
                input_mapping[i] = 0
            else:
                input_mapping[i] = 1
        # Build output mask
        output_mask = {'[0]': [0], '[1]': None}
        s.defineClassMapping(input_mapping)
        s.defineOutputMask(output_mask)
        s.setOptimizer(lr=stage1_lr)
        s.silence = False
        stage.append(s)

        outputs_list.append('loss_OnevsOne/output')

        logger.info('Built model %s/%s for classes %s in %0.5s seconds.' % (
            str(count + 1), str(n_classes), '(' + str(c) + ' vs All)',
            str(time.time() - t)))
        count += 1

    return [stage, outputs_list]


def build_Specific_OneVsOneECOC_Stage(pairs, input_shape, ds, lr, ecoc_version=2):
    """

    :param pairs:
    :param input_shape:
    :param ds:
    :param lr:
    :param ecoc_version:
    :return:
    """
    n_classes = len(ds.classes)

    stage = list()
    outputs_list = list()

    count = 0
    n_pairs = len(pairs)
    logger.info("Building " + str(n_pairs) + " classifiers...")

    for c in pairs:
        t = time.time()

        # Create each one_vs_one classifier of the intermediate stage
        if ecoc_version == 1:
            s = Stage(nInput=n_classes, nOutput=2, input_shape=input_shape,
                      output_shape=[2],
                      type='One_vs_One_Inception', silence=True)
        elif ecoc_version == 2:
            s = Stage(nInput=n_classes, nOutput=2, input_shape=input_shape,
                      output_shape=[2],
                      type='One_vs_One_Inception_v2', silence=True)
        # Build input mapping
        input_mapping = dict()
        for i in range(n_classes):
            if i in c:
                input_mapping[i] = c.index(i)
            else:
                input_mapping[i] = None
        # Build output mask
        # output_mask = {'[0]': [0], '[1]': None}
        s.defineClassMapping(input_mapping)
        # s.defineOutputMask(output_mask)
        s.setOptimizer(lr=lr)
        s.silence = False
        stage.append(s)

        outputs_list.append('loss_OnevsOne/output')

        logger.info('Built model %s/%s for classes %s = %s in %0.5s seconds.' % (
            str(count + 1), str(n_pairs), c, (ds.classes[c[0]], ds.classes[c[1]]),
            str(time.time() - t)))
        count += 1

    return [stage, outputs_list]


def build_Specific_OneVsOneVsRestECOC_Stage(pairs, input_shape, ds, lr,
                                            ecoc_version=2):
    """

    :param pairs:
    :param input_shape:
    :param ds:
    :param lr:
    :param ecoc_version:
    :return:
    """
    n_classes = len(ds.classes)

    stage = list()
    outputs_list = list()

    count = 0
    n_pairs = len(pairs)
    for c in pairs:
        t = time.time()

        # Create each one_vs_one classifier of the intermediate stage
        if ecoc_version == 1:
            s = Stage(nInput=n_classes, nOutput=3, input_shape=input_shape,
                      output_shape=[3],
                      type='One_vs_One_Inception', silence=True)
        elif ecoc_version == 2:
            s = Stage(nInput=n_classes, nOutput=3, input_shape=input_shape,
                      output_shape=[3],
                      type='One_vs_One_Inception_v2', silence=True)
        # Build input mapping
        input_mapping = dict()
        for i in range(n_classes):
            if i in c:
                input_mapping[i] = c.index(i)
            else:
                input_mapping[i] = 2
        # Build output mask
        # output_mask = {'[0]': [0], '[1]': None}
        s.defineClassMapping(input_mapping)
        # s.defineOutputMask(output_mask)
        s.setOptimizer(lr=lr)
        s.silence = False
        stage.append(s)

        outputs_list.append('loss_OnevsOne/output')

        logger.info('Built model %s/%s for classes %s = %s in %0.5s seconds.' % (
            str(count + 1), str(n_pairs), c, (ds.classes[c[0]], ds.classes[c[1]]),
            str(time.time() - t)))
        count += 1

    return [stage, outputs_list]


def build_Specific_OneVsOneECOC_loss_Stage(net, input_net, input_shape, classes,
                                           ecoc_version=3, pairs=None,
                                           functional_api=False, activations=None):
    """

    :param net:
    :param input_net:
    :param input_shape:
    :param classes:
    :param ecoc_version:
    :param pairs:
    :param functional_api:
    :param activations:
    :return:
    """
    from keras.layers.convolutional import ZeroPadding2D
    if activations is None:
        activations = ['softmax', 'softmax']
    n_classes = len(classes)
    if pairs is None:  # generate any possible combination of two classes
        pairs = tuple(itertools.combinations(range(n_classes), 2))

    outputs_list = list()
    n_pairs = len(pairs)
    ecoc_table = np.zeros((n_classes, n_pairs, 2))

    logger.info("Building " + str(n_pairs) + " OneVsOne structures...")

    for i, c in list(enumerate(pairs)):
        # t = time.time()

        # Insert 1s in the corresponding positions of the ecoc table
        ecoc_table[c[0], i, 0] = 1
        ecoc_table[c[1], i, 1] = 1

        # Create each one_vs_one classifier of the intermediate stage
        if not functional_api:
            if ecoc_version == 1:
                output_name = net.add_One_vs_One_Inception(input_net, input_shape, i,
                                                           nOutput=2,
                                                           activation=activations[0])
            elif ecoc_version == 2:
                output_name = net.add_One_vs_One_Inception_v2(input_net, input_shape,
                                                              i, nOutput=2,
                                                              activation=activations[
                                                                  0])
            else:
                raise NotImplementedError
        else:
            if ecoc_version == 1:
                output_name = net.add_One_vs_One_Inception_Functional(input_net,
                                                                      input_shape, i,
                                                                      nOutput=2,
                                                                      activation=activations[0])
            elif ecoc_version == 2:
                raise NotImplementedError()
            elif ecoc_version == 3 or ecoc_version == 4 or ecoc_version == 5 or ecoc_version == 6:
                if ecoc_version == 3:
                    nkernels = 16
                elif ecoc_version == 4:
                    nkernels = 64
                elif ecoc_version == 5:
                    nkernels = 128
                elif ecoc_version == 6:
                    nkernels = 256
                else:
                    raise NotImplementedError()
                if i == 0:
                    in_node = net.model.get_layer(input_net).output
                    padding_node = ZeroPadding2D(padding=(1, 1),
                                                 name='3x3/ecoc_padding')(in_node)
                output_name = net.add_One_vs_One_3x3_Functional(padding_node,
                                                                input_shape, i,
                                                                nkernels, nOutput=2,
                                                                activation=activations[0])
            elif ecoc_version == 7:
                if i == 0:
                    in_node = net.model.get_layer(input_net).output
                    padding_node = ZeroPadding2D(padding=(1, 1),
                                                 name='3x3/ecoc_padding')(in_node)
                output_name = net.add_One_vs_One_3x3_double_Functional(padding_node,
                                                                       input_shape,
                                                                       i, nOutput=2,
                                                                       activation=activations[0])
            else:
                raise NotImplementedError()
        outputs_list.append(output_name)

        # logger.info('Built model %s/%s for classes %s = %s in %0.5s seconds.'%(str(i+1),
        #  str(n_pairs), c, (classes[c[0]], classes[c[1]]), str(time.time()-t)))

    ecoc_table = np.reshape(ecoc_table, [n_classes, 2 * n_pairs])

    # Build final Softmax layer
    if not functional_api:
        output_names = net.add_One_vs_One_Merge(outputs_list, n_classes,
                                                activation=activations[1])
    else:
        output_names = net.add_One_vs_One_Merge_Functional(outputs_list, n_classes,
                                                           activation=activations[1])
    logger.info('Built ECOC merge layers.')

    return [ecoc_table, output_names]


def prepareECOCLossOutputs(net, ds, ecoc_table, input_name, output_names,
                           splits=None):
    """

    :param net:
    :param ds:
    :param ecoc_table:
    :param input_name:
    :param output_names:
    :param splits:
    :return:
    """
    if splits is None:
        splits = ['train', 'val', 'test']
    # Insert ecoc_table in net
    if 'additional_data' not in net.__dict__.keys():
        net.additional_data = dict()
    net.additional_data['ecoc_table'] = ecoc_table

    # Retrieve labels' id and images' id in dataset <- Necessary?
    # id_labels = ds.ids_outputs[ds.types_outputs.index('categorical')]
    id_labels_ecoc = 'labels_ecoc'

    # Insert ecoc-loss labels for each data split
    for s in splits:
        labels_ecoc = []
        # exec ('labels = ds.Y_' + s + '[id_labels]')
        y_split = getattr(ds, 'Y_' + s)
        labels = y_split[gt_id]
        n = len(labels)
        for i in range(n):
            labels_ecoc.append(ecoc_table[labels[i]])
        ds.setOutput(labels_ecoc, s, data_type='binary', id=id_labels_ecoc)

    # Set input and output mappings from dataset to network
    pos_images = ds.types_inputs.index('image')
    pos_labels = ds.types_outputs.index('categorical')
    pos_labels_ecoc = ds.types_outputs.index('binary')

    inputMapping = {input_name: pos_images}
    # inputMapping = {0: pos_images}
    net.setInputsMapping(inputMapping)

    outputMapping = {output_names[0]: pos_labels_ecoc, output_names[1]: pos_labels}
    # outputMapping = {0: pos_labels_ecoc, 1: pos_labels}
    net.setOutputsMapping(outputMapping, acc_output=output_names[1])


def loadGoogleNetForFood101(nClasses=101,
                            load_path='/media/HDD_2TB/CNN_MODELS/GoogleNet'):
    """

    :param nClasses:
    :param load_path:
    :return:
    """
    logger.info('Loading GoogLeNet...')

    # Build model (loading the previously converted Caffe's model)
    googLeNet = Stage(nClasses, nClasses, [224, 224, 3], [nClasses],
                      type='GoogleNet',
                      model_name='GoogleNet_Food101_retrained',
                      structure_path=load_path + '/Keras_model_structure.json',
                      weights_path=load_path + '/Keras_model_weights.h5')

    return googLeNet


def prepareGoogleNet_Food101(model_wrapper):
    """
    Prepares the GoogleNet model after its conversion from Caffe
    :param model_wrapper:
    :return:
    """
    # Remove unnecessary intermediate optimizers
    layers_to_delete = ['loss2/ave_pool', 'loss2/conv', 'loss2/relu_conv',
                        'loss2/fc_flatten', 'loss2/fc',
                        'loss2/relu_fc', 'loss2/drop_fc', 'loss2/classifier',
                        'output_loss2/loss',
                        'loss1/ave_pool', 'loss1/conv', 'loss1/relu_conv',
                        'loss1/fc_flatten', 'loss1/fc',
                        'loss1/relu_fc', 'loss1/drop_fc', 'loss1/classifier',
                        'output_loss1/loss']
    model_wrapper.removeLayers(layers_to_delete)
    model_wrapper.removeOutputs(['loss1/loss', 'loss2/loss'])


def prepareGoogleNet_Food101_ECOC_loss(model_wrapper):
    """
    Prepares the GoogleNet model for inserting an ECOC structure after removing the last part of the net
    :param model_wrapper:
    :return:
    """
    # Remove all last layers (from 'inception_5a' included)
    layers_to_delete = ['inception_5a/1x1', 'inception_5a/relu_1x1',
                        'inception_5a/3x3_reduce',
                        'inception_5a/relu_3x3_reduce',
                        'inception_5a/3x3_zeropadding', 'inception_5a/3x3',
                        'inception_5a/relu_3x3',
                        'inception_5a/5x5_reduce',
                        'inception_5a/relu_5x5_reduce',
                        'inception_5a/5x5_zeropadding', 'inception_5a/5x5',
                        'inception_5a/relu_5x5',
                        'inception_5a/pool_zeropadding', 'inception_5a/pool',
                        'inception_5a/pool_proj',
                        'inception_5a/relu_pool_proj', 'inception_5a/output',
                        'inception_5b/1x1',
                        'inception_5b/relu_1x1', 'inception_5b/3x3_reduce',
                        'inception_5b/relu_3x3_reduce',
                        'inception_5b/3x3_zeropadding', 'inception_5b/3x3',
                        'inception_5b/relu_3x3',
                        'inception_5b/5x5_reduce',
                        'inception_5b/relu_5x5_reduce',
                        'inception_5b/5x5_zeropadding', 'inception_5b/5x5',
                        'inception_5b/relu_5x5',
                        'inception_5b/pool_zeropadding', 'inception_5b/pool',
                        'inception_5b/pool_proj',
                        'inception_5b/relu_pool_proj',
                        'inception_5b/output', 'pool5/7x7_s1', 'pool5/drop_7x7_s1',
                        'loss3/classifier_foodrecognition_flatten',
                        'loss3/classifier_foodrecognition']
    [layers, params] = model_wrapper.removeLayers(copy.copy(layers_to_delete))
    # Remove softmax output
    model_wrapper.removeOutputs(['loss3/loss3'])

    return ['pool4/3x3_s2',
            [832, 7, 7]]  # returns the name of the last layer and its output shape
    # Adds a new output after the layer 'pool4/3x3_s2'
    # model_wrapper.model.add_output(name='pool4', input='pool4/3x3_s2')


def prepareGoogleNet_Food101_Stage1(model_wrapper):
    """
    Prepares the GoogleNet model for serving as the first Stage of a Staged_Netork
    :param model_wrapper:
    :return:
    """
    # Adds a new output after the layer 'pool4/3x3_s2'
    model_wrapper.model.add_output(name='pool4', input='pool4/3x3_s2')


def prepareGoogleNet_Stage2(stage1, stage2):
    """
    Removes the second part of the GoogleNet for inserting it into the second stage.
    :param stage1:
    :param stage2:
    :return:
    """
    # Remove all last layers (from 'inception_5a' included)
    layers_to_delete = ['inception_5a/1x1', 'inception_5a/relu_1x1',
                        'inception_5a/3x3_reduce',
                        'inception_5a/relu_3x3_reduce',
                        'inception_5a/3x3_zeropadding', 'inception_5a/3x3',
                        'inception_5a/relu_3x3',
                        'inception_5a/5x5_reduce',
                        'inception_5a/relu_5x5_reduce',
                        'inception_5a/5x5_zeropadding', 'inception_5a/5x5',
                        'inception_5a/relu_5x5',
                        'inception_5a/pool_zeropadding', 'inception_5a/pool',
                        'inception_5a/pool_proj',
                        'inception_5a/relu_pool_proj',
                        'inception_5a/output', 'inception_5b/1x1',
                        'inception_5b/relu_1x1', 'inception_5b/3x3_reduce',
                        'inception_5b/relu_3x3_reduce',
                        'inception_5b/3x3_zeropadding', 'inception_5b/3x3',
                        'inception_5b/relu_3x3',
                        'inception_5b/5x5_reduce',
                        'inception_5b/relu_5x5_reduce',
                        'inception_5b/5x5_zeropadding', 'inception_5b/5x5',
                        'inception_5b/relu_5x5',
                        'inception_5b/pool_zeropadding', 'inception_5b/pool',
                        'inception_5b/pool_proj',
                        'inception_5b/relu_pool_proj',
                        'inception_5b/output', 'pool5/7x7_s1', 'pool5/drop_7x7_s1',
                        'loss3/classifier_foodrecognition_flatten',
                        'loss3/classifier_foodrecognition', 'output_loss3/loss3']
    [layers, params] = stage1.removeLayers(copy.copy(layers_to_delete))
    # Remove softmax output
    stage1.removeOutputs(['loss3/loss3'])

    layers_to_delete_2 = ["conv1/7x7_s2_zeropadding", "conv1/7x7_s2",
                          "conv1/relu_7x7", "pool1/3x3_s2_zeropadding",
                          "pool1/3x3_s2", "pool1/norm1", "conv2/3x3_reduce",
                          "conv2/relu_3x3_reduce",
                          "conv2/3x3_zeropadding", "conv2/3x3", "conv2/relu_3x3",
                          "conv2/norm2",
                          "pool2/3x3_s2_zeropadding", "pool2/3x3_s2",
                          "inception_3a/1x1", "inception_3a/relu_1x1",
                          "inception_3a/3x3_reduce", "inception_3a/relu_3x3_reduce",
                          "inception_3a/3x3_zeropadding",
                          "inception_3a/3x3", "inception_3a/relu_3x3",
                          "inception_3a/5x5_reduce",
                          "inception_3a/relu_5x5_reduce",
                          "inception_3a/5x5_zeropadding", "inception_3a/5x5",
                          "inception_3a/relu_5x5", "inception_3a/pool_zeropadding",
                          "inception_3a/pool",
                          "inception_3a/pool_proj", "inception_3a/relu_pool_proj",
                          "inception_3a/output",
                          "inception_3b/1x1", "inception_3b/relu_1x1",
                          "inception_3b/3x3_reduce",
                          "inception_3b/relu_3x3_reduce",
                          "inception_3b/3x3_zeropadding", "inception_3b/3x3",
                          "inception_3b/relu_3x3", "inception_3b/5x5_reduce",
                          "inception_3b/relu_5x5_reduce",
                          "inception_3b/5x5_zeropadding", "inception_3b/5x5",
                          "inception_3b/relu_5x5",
                          "inception_3b/pool_zeropadding", "inception_3b/pool",
                          "inception_3b/pool_proj",
                          "inception_3b/relu_pool_proj", "inception_3b/output",
                          "pool3/3x3_s2_zeropadding",
                          "pool3/3x3_s2", "inception_4a/1x1",
                          "inception_4a/relu_1x1", "inception_4a/3x3_reduce",
                          "inception_4a/relu_3x3_reduce",
                          "inception_4a/3x3_zeropadding", "inception_4a/3x3",
                          "inception_4a/relu_3x3", "inception_4a/5x5_reduce",
                          "inception_4a/relu_5x5_reduce",
                          "inception_4a/5x5_zeropadding", "inception_4a/5x5",
                          "inception_4a/relu_5x5",
                          "inception_4a/pool_zeropadding", "inception_4a/pool",
                          "inception_4a/pool_proj",
                          "inception_4a/relu_pool_proj", "inception_4a/output",
                          "inception_4b/1x1",
                          "inception_4b/relu_1x1", "inception_4b/3x3_reduce",
                          "inception_4b/relu_3x3_reduce",
                          "inception_4b/3x3_zeropadding", "inception_4b/3x3",
                          "inception_4b/relu_3x3",
                          "inception_4b/5x5_reduce", "inception_4b/relu_5x5_reduce",
                          "inception_4b/5x5_zeropadding",
                          "inception_4b/5x5", "inception_4b/relu_5x5",
                          "inception_4b/pool_zeropadding",
                          "inception_4b/pool", "inception_4b/pool_proj",
                          "inception_4b/relu_pool_proj",
                          "inception_4b/output", "inception_4c/1x1",
                          "inception_4c/relu_1x1", "inception_4c/3x3_reduce",
                          "inception_4c/relu_3x3_reduce",
                          "inception_4c/3x3_zeropadding", "inception_4c/3x3",
                          "inception_4c/relu_3x3", "inception_4c/5x5_reduce",
                          "inception_4c/relu_5x5_reduce",
                          "inception_4c/5x5_zeropadding", "inception_4c/5x5",
                          "inception_4c/relu_5x5",
                          "inception_4c/pool_zeropadding", "inception_4c/pool",
                          "inception_4c/pool_proj",
                          "inception_4c/relu_pool_proj", "inception_4c/output",
                          "inception_4d/1x1",
                          "inception_4d/relu_1x1", "inception_4d/3x3_reduce",
                          "inception_4d/relu_3x3_reduce",
                          "inception_4d/3x3_zeropadding", "inception_4d/3x3",
                          "inception_4d/relu_3x3",
                          "inception_4d/5x5_reduce", "inception_4d/relu_5x5_reduce",
                          "inception_4d/5x5_zeropadding",
                          "inception_4d/5x5", "inception_4d/relu_5x5",
                          "inception_4d/pool_zeropadding",
                          "inception_4d/pool", "inception_4d/pool_proj",
                          "inception_4d/relu_pool_proj",
                          "inception_4d/output", "inception_4e/1x1",
                          "inception_4e/relu_1x1", "inception_4e/3x3_reduce",
                          "inception_4e/relu_3x3_reduce",
                          "inception_4e/3x3_zeropadding", "inception_4e/3x3",
                          "inception_4e/relu_3x3", "inception_4e/5x5_reduce",
                          "inception_4e/relu_5x5_reduce",
                          "inception_4e/5x5_zeropadding", "inception_4e/5x5",
                          "inception_4e/relu_5x5",
                          "inception_4e/pool_zeropadding", "inception_4e/pool",
                          "inception_4e/pool_proj",
                          "inception_4e/relu_pool_proj", "inception_4e/output",
                          "pool4/3x3_s2_zeropadding",
                          "pool4/3x3_s2"]

    # Remove initial layers
    [layers_, params_] = stage2.removeLayers(copy.copy(layers_to_delete_2))
    # Remove previous input
    stage2.removeInputs(['input_data'])
    # Add new input
    stage2.model.add_input(name='input_data', input_shape=(832, 7, 7))
    stage2.model.nodes[layers_to_delete[0]].previous = stage2.model.inputs[
        'input_data']
    """
    # Insert layers into stage
     stage2.model = Graph()
    # Input
     stage2.model.add_input(name='input_data', input_shape=(832,7,7))
     for l_name,l,p in zip(layers_to_delete, layers, params):
        stage2.model.namespace.add(l_name)
        stage2.model.nodes[l_name] = l
        stage2.model.node_config.append(p)
    #input = stage2.model.input # keep input
    # Connect first layer with input
     stage2.model.node_config[0]['input'] = 'input_data'
     stage2.model.nodes[layers_to_delete[0]].previous = stage2.model.inputs['input_data']
     stage2.model.input_config[0]['input_shape'] = [832,7,7]

    # Output
     stage2.model.add_output(name='loss3/loss3', input=layers_to_delete[-1])
    #stage2.model.add_output(name='loss3/loss3_', input=layers_to_delete[-1])
    #stage2.model.input = input # recover input
    """


def simplifyDataset(ds, id_classes, n_classes=50):
    """

    :param ds:
    :param id_classes:
    :param n_classes:
    :return:
    """
    logger.info("Simplifying %s from %d to %d classes." % (str(ds.name), len(ds.classes), n_classes))
    ds.classes[id_classes] = ds.classes[id_classes][:n_classes]

    id_labels = ds.ids_outputs[ds.types_outputs.index('categorical')]

    # reduce each data split
    for s in ['train', 'val', 'test']:
        kept_Y = dict()
        kept_X = dict()
        labels_set = getattr(ds, 'Y_' + s)[id_labels]
        for i, y in list(enumerate(labels_set)):
            if y < n_classes:
                for id_out in ds.ids_outputs:
                    # exec ('sample = ds.Y_' + s + '[id_out][i]')
                    y_split = getattr(ds, 'Y_' + s)
                    sample = y_split[id_out][i]
                    try:
                        kept_Y[id_out].append(sample)
                    except Exception:
                        kept_Y[id_out] = []
                        kept_Y[id_out].append(sample)
                for id_in in ds.ids_inputs:
                    # exec ('sample = ds.X_' + s + '[id_in][i]')
                    x_split = getattr(ds, 'X_' + s)
                    sample = x_split[id_in][i]
                    try:
                        kept_X[id_in].append(sample)
                    except Exception:
                        kept_X[id_in] = []
                        kept_X[id_in].append(sample)
        # exec ('ds.X_' + s + ' = copy.copy(kept_X)')
        # exec ('ds.Y_' + s + ' = copy.copy(kept_Y)')
        # exec ('ds.len_' + s + ' = len(kept_Y[id_labels])')
        setattr(ds, 'X_' + s, copy.copy(kept_X))
        setattr(ds, 'Y_' + s, copy.copy(kept_Y))
        setattr(ds, 'len_' + s, len(kept_Y[id_labels]))


def average_models(models, output_model, weights=None, custom_objects=None):
    from keras_wrapper.cnn_model import loadModel, saveModel
    if not isinstance(models, list):
        raise AssertionError('You must give a list of models to average.')
    if len(models) == 0:
        raise AssertionError('You provided an empty list of models to average!')

    model_weights = np.asarray([1. / len(models)] * len(models),
                               dtype=np.float32) if (weights is None) or (weights == []) else np.asarray(weights, dtype=np.float32)
    if len(model_weights) != len(models):
        raise AssertionError(
            'You must give a list of weights of the same size than the list of models.')
    loaded_models = [loadModel(m, -1, full_path=True, custom_objects=custom_objects) for m in models]

    # Check that all models are compatible
    if not all([hasattr(loaded_model, 'model') for loaded_model in loaded_models]):
        raise AssertionError('Not all models have the attribute "model".')
    if not (all([hasattr(loaded_model, 'model_init') for loaded_model in
                 loaded_models]) or all(
            [not hasattr(loaded_model, 'model_init') for loaded_model in
             loaded_models])):
        raise AssertionError('Not all models have the attribute "model_init".')

    if not (all([hasattr(loaded_model, 'model_next') for loaded_model in
                 loaded_models]) or all(
            [not hasattr(loaded_model, 'model_next') for loaded_model in
             loaded_models])):
        raise AssertionError('Not all models have the attribute "model_next".')

    # Check all layers are the same

    if not (all([[str(loaded_models[0].model.weights[i]) == str(loaded_model.model.weights[i]) for i in range(len(loaded_models[0].model.weights))] for loaded_model in loaded_models])):
        raise AssertionError('Not all models have the same weights!')

    if hasattr(loaded_models[0], 'model_init') and getattr(loaded_models[0], 'model_init') is not None:
        if not all([[str(loaded_models[0].model.weights[i]) == str(loaded_model.model.weights[i]) for i in range(len(loaded_models[0].model_init.weights))] for loaded_model in loaded_models]):
            raise AssertionError('Not all model_inits have the same weights!')

    if hasattr(loaded_models[0], 'model_next') and getattr(loaded_models[0], 'model_next') is not None:
        if not all([[str(loaded_models[0].model_next.weights[i]) == str(loaded_model.model_next.weights[i]) for i in range(len(loaded_models[0].model_next.weights))] for loaded_model in loaded_models]):
            raise AssertionError('Not all model_nexts have the same weights!')

    # Retrieve weights, weigh them and overwrite in model[0].
    current_weights = loaded_models[0].model.get_weights()
    loaded_models[0].model.set_weights(
        [current_weights[matrix_index] * model_weights[0] for matrix_index in range(len(current_weights))])
    # We have model_init
    if hasattr(loaded_models[0], 'model_init') and getattr(loaded_models[0], 'model_init') is not None:
        current_weights = loaded_models[0].model_init.get_weights()
        loaded_models[0].model_init.set_weights(
            [current_weights[matrix_index] * model_weights[0] for matrix_index in range(len(current_weights))])

    # We have model_next
    if hasattr(loaded_models[0], 'model_next') and getattr(loaded_models[0], 'model_next') is not None:
        current_weights = loaded_models[0].model_next.get_weights()
        loaded_models[0].model_next.set_weights(
            [current_weights[matrix_index] * model_weights[0] for matrix_index in range(len(current_weights))])

    # Weighted sum of all models
    for m in range(1, len(models)):
        current_weights = loaded_models[m].model.get_weights()
        prev_weights = loaded_models[0].model.get_weights()
        loaded_models[0].model.set_weights(
            [current_weights[matrix_index] * model_weights[m] + prev_weights[matrix_index] for matrix_index in range(len(current_weights))])

        # We have model_init
        if hasattr(loaded_models[0], 'model_init') and getattr(loaded_models[0], 'model_init') is not None:
            current_weights = loaded_models[m].model_init.get_weights()
            prev_weights = loaded_models[0].model_init.get_weights()
            loaded_models[0].model_init.set_weights(
                [current_weights[matrix_index] * model_weights[m] + prev_weights[matrix_index] for matrix_index in range(len(current_weights))])

        # We have model_next
        if hasattr(loaded_models[0], 'model_next') and getattr(loaded_models[0], 'model_next') is not None:
            current_weights = loaded_models[m].model_next.get_weights()
            prev_weights = loaded_models[0].model_next.get_weights()
            loaded_models[0].model_next.set_weights(
                [current_weights[matrix_index] * model_weights[m] + prev_weights[matrix_index] for matrix_index in range(len(current_weights))])

    # Save averaged model
    saveModel(loaded_models[0], -1, path=output_model, full_path=True,
              store_iter=False)


# Text-related utils
def one_hot_2_indices(preds, pad_sequences=True, verbose=0):
    """
    Converts a one-hot codification into a index-based one
    :param preds: Predictions codified as one-hot vectors.
    :param pad_sequences: Whether we should pad sequence or not
    :param verbose: Verbosity level, by default 0.
    :return: List of convertedpredictions
    """
    if verbose > 0:
        logger.info('Converting one hot prediction into indices...')
    preds = list(map(lambda x: np.argmax(x, axis=1), preds))
    if pad_sequences:
        preds = [pred[:sum([int(elem > 0) for elem in pred]) + 1] for pred in preds]
    return preds


def indices_2_one_hot(indices, n):
    """
    Converts a list of indices into one hot codification

    :param indices: list of indices
    :param n: integer. Size of the vocabulary
    :return: numpy array with shape (len(indices), n)
    """
    one_hot = np.zeros((len(indices), n), dtype=np.int8)
    for i in range(len(indices)):
        if indices[i] >= n:
            raise ValueError("Index out of bounds when converting to one hot")
        one_hot[i, indices[i]] = 1

    return one_hot


# From keras.utils.np_utils
def to_categorical(y, num_classes=None):
    """Converts a class vector (integers) to binary class matrix.

    E.g. for use with categorical_crossentropy.

    # Arguments
        y: class vector to be converted into a matrix
            (integers from 0 to num_classes).
        num_classes: total number of classes.

    # Returns
        A binary matrix representation of the input.
    """
    y = np.array(y, dtype='int')
    input_shape = y.shape
    if input_shape and input_shape[-1] == 1 and len(input_shape) > 1:
        input_shape = tuple(input_shape[:-1])
    y = y.ravel()
    if not num_classes:
        num_classes = np.max(y) + 1
    n = y.shape[0]
    categorical = np.zeros((n, num_classes))
    categorical[np.arange(n), y] = 1
    output_shape = input_shape + (num_classes,)
    categorical = np.reshape(categorical, output_shape)
    return categorical


def categorical_probas_to_classes(p):
    return np.argmax(p, axis=1)


# ------------------------------------------------------- #
#       DECODING FUNCTIONS
#           Functions for decoding predictions
# ------------------------------------------------------- #

def decode_predictions_one_hot(preds, index2word, pad_sequences=True, verbose=0):
    """
    Decodes predictions following a one-hot codification.
    :param preds: Predictions codified as one-hot vectors.
    :param index2word: Mapping from word indices into word characters.
    :param verbose: Verbosity level, by default 0.
    :return: List of decoded predictions
    """
    if verbose > 0:
        logger.info('Decoding one hot prediction ...')
    preds = list(map(lambda prediction: np.argmax(prediction, axis=1), preds))
    PAD = '<pad>'
    flattened_answer_pred = [list(map(lambda index: index2word[index], pred)) for
                             pred in preds]
    answer_pred_matrix = np.asarray(flattened_answer_pred)
    answer_pred = []

    for a_no in answer_pred_matrix:
        end_token_pos = [j for j, x in list(enumerate(a_no)) if x == PAD]
        end_token_pos = None if len(end_token_pos) == 0 or not pad_sequences else end_token_pos[0]
        a_no = [a.decode('utf-8') if isinstance(a,
                                                str) and sys.version_info.major == 2 else a
                for a in a_no]
        tmp = u' '.join(a_no[:end_token_pos])
        answer_pred.append(tmp)
    return answer_pred


def decode_predictions(preds, temperature, index2word, sampling_type, verbose=0):
    """
    Decodes predictions
    :param preds: Predictions codified as the output of a softmax activation function.
    :param temperature: Temperature for sampling.
    :param index2word: Mapping from word indices into word characters.
    :param sampling_type: 'max_likelihood' or 'multinomial'.
    :param verbose: Verbosity level, by default 0.
    :return: List of decoded predictions.
    """

    if verbose > 0:
        logger.info('Decoding prediction ...')
    flattened_preds = preds.reshape(-1, preds.shape[-1])
    flattened_answer_pred = list(map(lambda index: index2word[index], sampling(scores=flattened_preds,
                                                                               sampling_type=sampling_type,
                                                                               temperature=temperature)))
    answer_pred_matrix = np.asarray(flattened_answer_pred).reshape(preds.shape[:-1])

    answer_pred = []
    EOS = '<eos>'
    PAD = '<pad>'

    for a_no in answer_pred_matrix:
        if len(a_no.shape) > 1:  # only process word by word if our prediction has more than one output
            init_token_pos = 0
            end_token_pos = [j for j, x in list(enumerate(a_no)) if x == EOS or x == PAD]
            end_token_pos = None if len(end_token_pos) == 0 else end_token_pos[0]
            a_no = [a.decode('utf-8') if isinstance(a, str) and sys.version_info.major == 2 else a
                    for a in a_no]
            tmp = u' '.join(a_no[init_token_pos:end_token_pos])
        else:
            tmp = a_no[:-1]
        answer_pred.append(tmp)
    return answer_pred


def decode_categorical(preds, index2word, verbose=0):
    """
    Decodes predictions
    :param preds: Predictions codified as the output of a softmax activation function.
    :param index2word: Mapping from word indices into word characters.
    :return: List of decoded predictions.
    """

    if verbose > 0:
        logger.info('Decoding prediction ...')

    word_indices = categorical_probas_to_classes(preds)
    return [index2word.get(word) for word in word_indices]


def decode_multilabel(preds, index2word, min_val=0.5, get_probs=False, verbose=0):
    """
    Decodes predictions
    :param preds: Predictions codified as the output of a softmax activation function.
    :param index2word: Mapping from word indices into word characters.
    :param min_val: Minimum value needed for considering a positive prediction.
    :param get_probs: additionally return probability for each predicted label
    :param verbose: Verbosity level, by default 0.
    :return: List of decoded predictions.
    """

    if verbose > 0:
        logger.info('Decoding prediction ...')

    answer_pred = []
    probs_pred = []
    for pred in preds:
        current_pred = []
        current_probs = []
        for ind, word in list(enumerate(pred)):
            if word >= min_val:
                current_pred.append(index2word[ind])
                current_probs.append(word)
        answer_pred.append(current_pred)
        probs_pred.append(current_probs)

    if get_probs:
        return answer_pred, probs_pred
    else:
        return answer_pred


def replace_unknown_words(src_word_seq, trg_word_seq, hard_alignment, unk_symbol,
                          glossary=None, heuristic=0, mapping=None, verbose=0):
    """
    Replaces unknown words from the target sentence according to some heuristic.
    Borrowed from: https://github.com/sebastien-j/LV_groundhog/blob/master/experiments/nmt/replace_UNK.py
    :param src_word_seq: Source sentence words
    :param trg_word_seq: Hypothesis words
    :param hard_alignment: Target-Source alignments
    :param glossary: Hard-coded substitutions.
    :param unk_symbol: Symbol in trg_word_seq to replace
    :param heuristic: Heuristic (0, 1, 2)
    :param mapping: External alignment dictionary
    :param verbose: Verbosity level
    :return: trg_word_seq with replaced unknown words
    """
    trans_words = trg_word_seq
    new_trans_words = []
    for j in range(len(trans_words)):
        current_word = trans_words[j]
        if glossary is not None and glossary.get(
                src_word_seq[hard_alignment[j]]) is not None:
            current_word = glossary.get(src_word_seq[hard_alignment[j]])
            new_trans_words.append(current_word)
        elif current_word == unk_symbol:
            current_src = src_word_seq[hard_alignment[j]]
            if isinstance(current_src, str) and sys.version_info.major == 2:
                current_src = current_src.decode('utf-8')
            if heuristic == 0:  # Copy (ok when training with large vocabularies on en->fr, en->de)
                new_trans_words.append(current_src)
            elif heuristic == 1:
                # Use the most likely translation (with t-table). If not found, copy the source word.
                # Ok for small vocabulary (~30k) models
                if mapping.get(current_src) is not None:
                    new_trans_words.append(mapping[current_src])
                else:
                    new_trans_words.append(current_src)
            elif heuristic == 2:
                # Use t-table if the source word starts with a lowercase letter. Otherwise copy
                # Sometimes works better than other heuristics
                if mapping.get(current_src) is not None and current_src[0].islower():
                    new_trans_words.append(mapping[current_src])
                else:
                    new_trans_words.append(current_src)
        else:
            new_trans_words.append(current_word)
    return new_trans_words


def decode_predictions_beam_search(preds, index2word, glossary=None, alphas=None,
                                   heuristic=0,
                                   x_text=None, unk_symbol='<unk>',
                                   pad_sequences=False,
                                   mapping=None, verbose=0):
    """
    Decodes predictions from the BeamSearch method.

    :param preds: Predictions codified as word indices.
    :param index2word: Mapping from word indices into word characters.
    :param alphas: Attention model weights: Float matrix with shape (I, J) (I: number of target items; J: number of source items).
    :param heuristic: Replace unknown words heuristic (0, 1 or 2)
    :param x_text: Source text (for unk replacement)
    :param unk_symbol: Unknown words symbol
    :param pad_sequences: Whether we should make a zero-pad on the input sequence.
    :param mapping: Source-target dictionary (for unk_replace heuristics 1 and 2)
    :param verbose: Verbosity level, by default 0.
    :return: List of decoded predictions
    """
    if verbose > 0:
        logger.info('Decoding beam search prediction ...')

    if alphas is not None:
        if x_text is None:
            raise AssertionError('When using POS_UNK, you must provide the input '
                                 'text to decode_predictions_beam_search!')
        if verbose > 0:
            logger.info('Using heuristic %d' % heuristic)
    if pad_sequences:
        preds = [pred[:sum([int(elem > 0) for elem in pred]) + 1] for pred in preds]
    flattened_predictions = [list(map(lambda x: index2word[x], pred)) for pred in
                             preds]
    final_predictions = []

    if alphas is not None:
        x_text = list(map(lambda x: x.split(), x_text))
        hard_alignments = list(
            map(lambda alignment, x_sentence: np.argmax(
                alignment[:, :max(1, len(x_sentence))], axis=1), alphas,
                x_text))

        for i, a_no in list(enumerate(flattened_predictions)):
            if unk_symbol in a_no or glossary is not None:
                a_no = replace_unknown_words(x_text[i],
                                             a_no,
                                             hard_alignments[i],
                                             unk_symbol,
                                             glossary=glossary,
                                             heuristic=heuristic,
                                             mapping=mapping,
                                             verbose=verbose)
            a_no = [a.decode('utf-8') if isinstance(a,
                                                    str) and sys.version_info.major == 2 else a
                    for a in a_no]
            tmp = u' '.join(a_no[:-1])
            final_predictions.append(tmp)
    else:
        for a_no in flattened_predictions:
            a_no = [a.decode('utf-8') if isinstance(a,
                                                    str) and sys.version_info.major == 2 else a
                    for a in a_no]
            tmp = u' '.join(a_no[:-1])
            final_predictions.append(tmp)
    return final_predictions


def sampling(scores, sampling_type='max_likelihood', temperature=1.0):
    """
    Sampling words (each sample is drawn from a categorical distribution).
    Or picks up words that maximize the likelihood.
    :param scores: array of size #samples x #classes;
    every entry determines a score for sample i having class j
    :param sampling_type:
    :param temperature: Predictions temperature. The higher, the flatter probabilities. Hence more random outputs.
    :return: set of indices chosen as output, a vector of size #samples
    """
    if isinstance(scores, dict):
        scores = scores['output']

    if sampling_type == 'multinomial':
        preds = np.asarray(scores).astype('float64')
        preds = np.log(preds) / temperature
        exp_preds = np.exp(preds)
        preds = exp_preds / np.sum(exp_preds)
        probas = np.random.multinomial(1, preds, 1)
        return np.argmax(probas)
    elif sampling_type == 'max_likelihood':
        return np.argmax(scores, axis=-1)
    else:
        raise NotImplementedError()


# Data structures-related utils
def flatten_list_of_lists(list_of_lists):
    """
    Flattens a list of lists
    :param list_of_lists: List of lists
    :return: Flatten list of lists
    """
    return [item for sublist in list_of_lists for item in sublist]


def flatten(l):
    """
    Flatten a list (more general than flatten_list_of_lists, but also more inefficient
    :param l:
    :return:
    """
    if not l:
        return l
    return flatten(l[0]) + (flatten(l[1:]) if len(l) > 1 else []) if isinstance(l,
                                                                                list) else [
        l]


def key_with_max_val(d):
    """ a) create a list of the dict's keys and values;
        b) return the key with the max value"""

    d = dict((k, v) for k, v in iteritems(d) if isinstance(v, (int, float, complex)))
    v = list(d.values())
    k = list(d.keys())
    if d == {}:
        return -1
    else:
        return k[v.index(max(v))]


def print_dict(d, header=''):
    """
    Formats a dictionary for printing.
    :param d: Dictionary to print.
    :return: String containing the formatted dictionary.
    """
    obj_str = str(header) + '{ \n\t'
    obj_str += "\n\t".join([str(key) + ": " + str(d[key]) for key in sorted(d.keys())])
    obj_str += '\n'
    obj_str += '}'
    return obj_str



from keras.layers.noise import GaussianNoise
from keras.layers.advanced_activations import ChannelWisePReLU as PReLU
from keras.layers.normalization import BatchNormalization, L2_norm, L1_norm
from keras.regularizers import l2
from keras.layers.core import Dropout, Lambda


def Regularize(layer, params,
               shared_layers=False,
               name='',
               apply_noise=True,
               apply_batch_normalization=True,
               apply_prelu=True,
               apply_dropout=True,
               apply_l1=True,
               apply_l2=True):
    """
    Apply the regularization specified in parameters to the layer
    :param layer: Layer to regularize
    :param params: Params specifying the regularizations to apply
    :param shared_layers: Boolean indicating if we want to get the used layers for applying to a shared-layers model.
    :param name: Name prepended to regularizer layer
    :param apply_noise: If False, noise won't be applied, independently of params
    :param apply_dropout: If False, dropout won't be applied, independently of params
    :param apply_prelu: If False, prelu won't be applied, independently of params
    :param apply_batch_normalization: If False, batch normalization won't be applied, independently of params
    :param apply_l1: If False, l1 normalization won't be applied, independently of params
    :param apply_l2: If False, l2 normalization won't be applied, independently of params
    :return: Regularized layer
    """
    shared_layers_list = []

    if apply_noise and params.get('USE_NOISE', False):
        shared_layers_list.append(GaussianNoise(params.get('NOISE_AMOUNT', 0.01), name=name + '_gaussian_noise'))

    if apply_batch_normalization and params.get('USE_BATCH_NORMALIZATION', False):
        if params.get('WEIGHT_DECAY'):
            l2_gamma_reg = l2(params['WEIGHT_DECAY'])
            l2_beta_reg = l2(params['WEIGHT_DECAY'])
        else:
            l2_gamma_reg = None
            l2_beta_reg = None

        bn_mode = params.get('BATCH_NORMALIZATION_MODE', 0)

        shared_layers_list.append(BatchNormalization(mode=bn_mode,
                                                     gamma_regularizer=l2_gamma_reg,
                                                     beta_regularizer=l2_beta_reg,
                                                     name=name + '_batch_normalization'))

    if apply_prelu and params.get('USE_PRELU', False):
        shared_layers_list.append(PReLU(name=name + '_PReLU'))

    if apply_dropout and params.get('DROPOUT_P', 0) > 0:
        shared_layers_list.append(Dropout(params.get('DROPOUT_P', 0.5), name=name + '_dropout'))

    if apply_l1 and params.get('USE_L1', False):
        shared_layers_list.append(Lambda(L1_norm, name=name + '_L1_norm'))

    if apply_l2 and params.get('USE_L2', False):
        shared_layers_list.append(Lambda(L2_norm, name=name + '_L2_norm'))

    # Apply all the previously built shared layers
    for l in shared_layers_list:
        layer = l(layer)
    result = layer

    # Return result or shared layers too
    if shared_layers:
        return result, shared_layers_list
    return result



# -*- coding: utf-8 -*-
from __future__ import print_function
from six import iteritems
from builtins import map, zip
import json
import logging
logging.basicConfig(level=logging.INFO, format='[%(asctime)s] %(message)s', datefmt='%d/%m/%Y %H:%M:%S')
logger = logging.getLogger(__name__)

from keras_wrapper.extra.localization_utilities import *


# EVALUATION FUNCTIONS SELECTOR

def get_coco_score(pred_list, verbose, extra_vars, split):
    """
    COCO challenge metrics
    :param pred_list: dictionary of hypothesis sentences (id, sentence)
    :param verbose: if greater than 0 the metric measures are printed out
    :param extra_vars: extra variables, here are:
            extra_vars['references'] - dict mapping sample indices to list with all valid captions (id, [sentences])
            extra_vars['tokenize_f'] - tokenization function used during model training (used again for validation)
            extra_vars['detokenize_f'] - detokenization function used during model training (used again for validation)
            extra_vars['tokenize_hypotheses'] - Whether tokenize or not the hypotheses during evaluation
    :param split: split on which we are evaluating
    :return: Dictionary with the coco scores
    """
    from pycocoevalcap.bleu.bleu import Bleu
    from pycocoevalcap.meteor.meteor import Meteor
    from pycocoevalcap.meteor import accepted_langs
    from pycocoevalcap.cider.cider import Cider
    from pycocoevalcap.rouge.rouge import Rouge
    from pycocoevalcap.ter.ter import Ter

    gts = extra_vars[split]['references']
    if extra_vars.get('tokenize_hypotheses', False):
        hypo = {idx: list(map(extra_vars['tokenize_f'], [lines.strip()])) for (idx, lines) in
                list(enumerate(pred_list))}
    else:
        hypo = {idx: [lines.strip()] for (idx, lines) in list(enumerate(pred_list))}

    # Tokenize refereces if needed
    if extra_vars.get('tokenize_references', False):
        refs = {idx: list(map(extra_vars['tokenize_f'], gts[idx])) for idx in list(gts)}
    else:
        refs = gts

    # Detokenize references if needed.
    if extra_vars.get('apply_detokenization', False):
        refs = {idx: list(map(extra_vars['detokenize_f'], refs[idx])) for idx in refs}

    scorers = [
        (Bleu(4), ["Bleu_1", "Bleu_2", "Bleu_3", "Bleu_4"]),
        (Ter(), "TER"),
        (Rouge(), "ROUGE_L"),
        (Cider(), "CIDEr")
    ]
    if extra_vars.get('language', 'en') in accepted_langs:
        scorers.append((Meteor(language=extra_vars.get('language', 'en')), "METEOR"))

    final_scores = {}
    for scorer, method in scorers:
        score, _ = scorer.compute_score(refs, hypo)
        if isinstance(score, list):
            for m, s in list(zip(method, score)):
                final_scores[m] = s
        else:
            final_scores[method] = score

    if verbose > 0:
        logger.info('Computing coco scores on the %s split...' % split)
    for metric in sorted(final_scores):
        value = final_scores[metric]
        logger.info(metric + ': ' + str(value))

    return final_scores


def eval_vqa(pred_list, verbose, extra_vars, split):
    """
    VQA challenge metrics
    :param pred_list: dictionary of hypothesis sentences (id, sentence)
    :param verbose: if greater than 0 the metric measures are printed out
    :param extra_vars: extra variables
            extra_vars['quesFile'] - path to the .json file where the questions are stored
            extra_vars['annFile'] - path to the .json file where the annotated answers are stored
            extra_vars['question_ids'] - question identifiers
    :param split: split on which we are evaluating
    :return: Dictionary of VQA accuracies
    """
    import datetime
    import os
    from pycocoevalcap.vqa import vqaEval, visual_qa
    from read_write import list2vqa

    quesFile = extra_vars[split]['quesFile']
    annFile = extra_vars[split]['annFile']

    # create temporal resFile
    resFile = 'tmp_res_file_' + str(datetime.datetime.now()) + '.json'
    list2vqa(resFile, pred_list, extra_vars[split]['question_ids'])

    # create vqa object and vqaRes object
    vqa_ = visual_qa.VQA(annFile, quesFile)
    vqaRes = vqa_.loadRes(resFile, quesFile)
    vqaEval_ = vqaEval.VQAEval(vqa_, vqaRes, n=2)  # n is precision of accuracy (number of places after decimal), default is 2
    vqaEval_.evaluate()
    os.remove(resFile)  # remove temporal file

    # get results
    acc_overall = vqaEval_.accuracy['overall']
    acc_yes_no = vqaEval_.accuracy['perAnswerType']['yes/no']
    acc_number = vqaEval_.accuracy['perAnswerType']['number']
    acc_other = vqaEval_.accuracy['perAnswerType']['other']
    # acc_per_class = vqaEval_.accuracy['perAnswerClass']
    # acc_class_normalized = vqaEval_.accuracy['classNormalizedOverall']

    if verbose > 0:
        logger.info('VQA Metric: Accuracy yes/no is {0}, other is {1}, number is {2}, overall is {3}'.
                    format(acc_yes_no, acc_other, acc_number, acc_overall))
    return {'overall accuracy': acc_overall,
            'yes/no accuracy': acc_yes_no,
            'number accuracy': acc_number,
            'other accuracy': acc_other}


def multilabel_metrics(pred_list, verbose, extra_vars, split):
    """
    Multiclass classification metrics. see multilabel ranking metrics in sklearn library for more info:
        http://scikit-learn.org/stable/modules/model_evaluation.html#multilabel-ranking-metrics

    :param pred_list: dictionary of hypothesis sentences
    :param verbose: if greater than 0 the metric measures are printed out
    :param extra_vars: extra variables
                        extra_vars['word2idx'] - dictionary mapping from words to indices
                        extra_vars['references'] - list of GT labels
    :param split: split on which we are evaluating
    :return: Dictionary of multilabel metrics
    """
    from sklearn import metrics as sklearn_metrics

    word2idx = extra_vars[split]['word2idx']

    # check if an additional dictionary matching raw to basic and general labels is provided
    # in that case a more general evaluation will be considered
    raw2basic = extra_vars[split].get('raw2basic', None)
    if raw2basic is not None:
        logger.info('Applying general evaluation with raw2basic dictionary.')

    if raw2basic is None:
        n_classes = len(word2idx)
    else:
        basic_values = set(raw2basic.values())
        n_classes = len(basic_values)
    n_samples = len(pred_list)

    # Create prediction matrix
    y_pred = np.zeros((n_samples, n_classes))
    for i_s, sample in list(enumerate(pred_list)):
        for word in sample:
            if raw2basic is None:
                y_pred[i_s, word2idx[word]] = 1
            else:
                word = word.strip()
                y_pred[i_s, raw2basic[word]] = 1

    # Prepare GT
    gt_list = extra_vars[split]['references']

    if raw2basic is None:
        y_gt = np.array(gt_list)
    else:
        idx2word = {v: k for k, v in iteritems(word2idx)}
        y_gt = np.zeros((n_samples, n_classes))
        for i_s, sample in list(enumerate(gt_list)):
            for raw_idx, is_active in list(enumerate(sample)):
                if is_active:
                    word = idx2word[raw_idx].strip()
                    y_gt[i_s, raw2basic[word]] = 1

    # Compute Coverage Error
    coverr = sklearn_metrics.coverage_error(y_gt, y_pred)
    # Compute Label Ranking AvgPrec
    avgprec = sklearn_metrics.label_ranking_average_precision_score(y_gt, y_pred)
    # Compute Label Ranking Loss
    rankloss = sklearn_metrics.label_ranking_loss(y_gt, y_pred)
    # Compute Precision, Recall and F1 score
    precision, recall, f1, _ = sklearn_metrics.precision_recall_fscore_support(y_gt, y_pred, average='micro')

    if verbose > 0:
        logger.info(
            '"coverage_error" (best: avg labels per sample = %f): %f' % (float(np.sum(y_gt)) / float(n_samples), coverr))
        logger.info('Label Ranking "average_precision" (best: 1.0): %f' % avgprec)
        logger.info('Label "ranking_loss" (best: 0.0): %f' % rankloss)
        logger.info('precision: %f' % precision)
        logger.info('recall: %f' % recall)
        logger.info('f1: %f' % f1)

    return {'coverage_error': coverr,
            'average_precision': avgprec,
            'ranking_loss': rankloss,
            'precision': precision,
            'recall': recall,
            'f1': f1}


import numpy as np


def multiclass_metrics(pred_list, verbose, extra_vars, split):
    """
    Multiclass classification metrics. See multilabel ranking metrics in sklearn library for more info:
        http://scikit-learn.org/stable/modules/model_evaluation.html#multilabel-ranking-metrics
    :param pred_list: list of predictions
    :param verbose: if greater than 0 the metric measures are printed out
    :param extra_vars: dictionary extra variables. Must contain:
                - ['n_classes'] with the total number of existent classes
                - [split]['references'] with the GT values corresponding to each sample of the current data split
    :param split: split of the data where we are applying the evaluation
    :return: dictionary of multiclass metrics
    """
    from sklearn import metrics as sklearn_metrics

    n_classes = extra_vars['n_classes']

    n_samples = len(pred_list)
    logger.info("---# of samples: " + str(n_samples))
    gt_list = extra_vars[split]['references']
    pred_class_list = [np.argmax(sample_score) for sample_score in pred_list]
    # Create prediction matrix
    y_pred = np.zeros((n_samples, n_classes))
    y_gt = np.zeros((n_samples, n_classes))
    for i_s, pred_class in list(enumerate(pred_class_list)):
        y_pred[i_s, pred_class] = 1
    try:
        values_gt = list(gt_list.values())
    except AttributeError:
        values_gt = gt_list

    counts_per_class = np.zeros((n_classes,))
    for i_s, gt_class in list(enumerate(values_gt)):
        y_gt[i_s, gt_class] = 1
        counts_per_class[gt_class] += 1

    # Apply balanced accuracy per class
    inverse_counts_per_class = [sum(counts_per_class) - c_i for c_i in counts_per_class]
    weights_per_class = [float(c_i) / sum(inverse_counts_per_class) for c_i in inverse_counts_per_class]
    sample_weights = np.zeros((n_samples,))
    for i_s, gt_class in list(enumerate(values_gt)):
        sample_weights[i_s] = weights_per_class[gt_class]

    # Compute accuracy
    accuracy = sklearn_metrics.accuracy_score(y_gt, y_pred)
    accuracy_balanced = sklearn_metrics.accuracy_score(y_gt, y_pred, sample_weight=sample_weights)
    # Compute Precision, Recall and F1 score
    avrg = extra_vars.get('average_mode', 'macro')
    precision, recall, f1, _ = sklearn_metrics.precision_recall_fscore_support(y_gt, y_pred, average=avrg)
    # Compute Confusion Matrix
    cf = sklearn_metrics.confusion_matrix(np.argmax(y_gt, -1), np.argmax(y_pred, -1))
    identity = np.identity(n_classes)
    neg_identity = 1 - identity
    # Compute TP, FP and FN from Confusion Matrix.
    tp = np.diag(cf)
    fp = np.sum(cf * neg_identity, axis=1)
    fn = np.sum(cf * neg_identity, axis=0)
    # Compute precision and recall per class
    condition_positive = tp + fn
    pred_condition_positive = tp + fp
    precision_per_class = tp / pred_condition_positive
    recall_per_class = tp / condition_positive
    # Compute top 5 fp classes
    top5_fps = np.argpartition(cf * neg_identity, -5)[:, -5:][:, ::-1]
    # Compute top 5 accuracy
    arg_top5_pred = np.argpartition(pred_list, -5)[:, -5:]
    arg_gt = np.argmax(y_gt, -1)
    top5_acc = np.mean(np.max(arg_top5_pred == np.repeat(np.expand_dims(arg_gt, -1), 5, -1), -1))

    if verbose > 0:
        logger.info('Top5 Accuracy: %f' % top5_acc)
        logger.info('accuracy: %f' % accuracy, )
        logger.info('balanced_accuracy: %f' % accuracy_balanced)
        logger.info('precision: ' + str(precision))
        logger.info('recall: ' + str(recall))
        logger.info('f1: ' + str(f1))

    return {'accuracy': accuracy,
            'balanced_accuracy': accuracy_balanced,
            'precision': precision,
            'recall': recall,
            'f1': f1,
            'top5_acc': top5_acc,
            'precision_per_class': list(precision_per_class),
            'recall_per_class': list(recall_per_class),
            'top5_fps': list(top5_fps)}


def semantic_segmentation_accuracy(pred_list, verbose, extra_vars, split):
    """
    Semantic Segmentation Accuracy metric

    :param pred_list: list of predictions
    :param verbose: if greater than 0 the metric measures are printed out
    :param extra_vars: dictionary extra variables. Must contain:
                       - ['n_classes'] with the total number of existent classes
                       - [split]['references'] with the GT values corresponding to each sample of the current data split
                       - [discard_classes] with the classes not taken into account in the performance evaluation
    :param split: split of the data where we are applying the evaluation
    :return: dictionary containing the semantic global accuracy
    """
    from sklearn import metrics as sklearn_metrics

    n_classes = extra_vars['n_classes']
    gt_list = extra_vars[split]['references']
    discard_classes = extra_vars['discard_classes']  # [11]

    pred_class_list = []
    for sample_score in pred_list:
        pred_class_list += list(np.argmax(sample_score, axis=1))
    n_samples = len(pred_class_list)

    values_gt = []
    for gt in gt_list:
        values_gt += list(gt)

    # Create ground truth and prediction matrices
    y_gt = np.zeros((n_samples, n_classes))
    y_pred = np.zeros((n_samples, n_classes))

    ind_i = 0
    for _, (gt_class, pred_class) in list(enumerate(zip(values_gt, pred_class_list))):
        if not any([d == gt_class for d in discard_classes]):
            y_pred[ind_i, pred_class] = 1
            y_gt[ind_i, gt_class] = 1
            ind_i += 1
    n_samples = ind_i

    # Cut to real n_samples size
    y_gt = y_gt[:n_samples]
    y_pred = y_pred[:n_samples]

    # Compute Coverage Error
    accuracy = sklearn_metrics.accuracy_score(y_gt, y_pred)
    if verbose > 0:
        logger.info('Accuracy: %f' % accuracy)

    return {'semantic global accuracy': accuracy}


def semantic_segmentation_meaniou(pred_list, verbose, extra_vars, split):
    """
    Semantic Segmentation Mean IoU metric

    :param pred_list: list of predictions
    :param verbose: if greater than 0 the metric measures are printed out
    :param extra_vars: dictionary extra variables. Must contain:
                    - ['n_classes'] with the total number of existent classes
                    - [split]['references'] with the GT values corresponding to each sample of the current data split
                    - [discard_classes] with the classes not taken into account in the performance evaluation
    :param split: split of the data where we are applying the evaluation
    :return: dictionary containing the mean IoU and the semantic global accuracy metrics
    """
    from sklearn import metrics as sklearn_metrics

    n_classes = extra_vars['n_classes']
    gt_list = extra_vars[split]['references']
    discard_classes = extra_vars['discard_classes']

    pred_class_list = []
    for sample_score in pred_list:
        pred_class_list += list(np.argmax(sample_score, axis=1))
    n_samples = len(pred_class_list)

    values_gt = []
    for gt in gt_list:
        values_gt += list(gt)

    # Create ground truth and prediction matrices
    y_gt = np.zeros((n_samples,))
    y_pred = np.zeros((n_samples,))

    ind_i = 0
    for _, (gt_class, pred_class) in list(enumerate(zip(values_gt, pred_class_list))):
        if not any([d == gt_class for d in discard_classes]):
            y_gt[ind_i] = gt_class
            y_pred[ind_i] = pred_class
            ind_i += 1
    n_samples = ind_i

    # Cut to real n_samples size
    y_gt = y_gt[:n_samples]
    y_pred = y_pred[:n_samples]

    # Computer mean IoU (Jaccard index) and pixel accuracy
    cm = sklearn_metrics.confusion_matrix(y_gt, y_pred)
    cm_t = cm.transpose()

    inter = np.zeros(n_classes - len(discard_classes))
    union = np.zeros(n_classes - len(discard_classes))
    ind_i = 0

    for l in range(0, n_classes):
        if not any([d == l for d in discard_classes]):
            tp = cm[l][l]
            fn = np.sum(cm[l]) - tp
            fp = np.sum(cm_t[l]) - tp
            inter[ind_i] = float(tp)
            union[ind_i] = float(tp + fp + fn)
            ind_i += 1

    mean_iou = np.mean(inter / union)
    acc = np.sum(inter) / np.sum(cm)

    if verbose > 0:
        logger.info('Mean IoU: %f' % mean_iou)
        logger.info('Accuracy: %f' % float(acc))

    return {'mean IoU': mean_iou, 'semantic global accuracy': acc}


def averagePrecision(pred_list, verbose, extra_vars, split):
    """
    Computes a Precision-Recall curve and its associated mAP score given a set of precalculated reports.
    The parameter "report_all" must include the following information for each sample:
        [predicted_bboxes, predicted_Y, predicted_scores, GT_bboxes, GT_Y]
    The parameter 'threshods' defines the range of min_prediction_scores to be tested for computing the
    precision-recall curve.

    :param pred_list: list of hypotheses with the following info [predicted_bboxes, predicted_Y, predicted_scores]
    :param verbose: if greater than 0 the metric measures are printed out
    :param extra_vars: extra variables, here are:
                extra_vars[split]['references'] - list of samples with the following info [GT_bboxes, GT_Y]
                extra_vars['n_classes'] - number of classes in the detection task
    :param split: split of the data where we are applying the evaluation
    :return: dictionary containing the aforementioned metrics.
    """

    thresholds = np.arange(0, 1, 0.1)
    fixedIoU = 0.5

    gts = extra_vars[split]['references']
    n_classes = extra_vars['n_classes']

    n_samples = len(pred_list)
    n_thresholds = len(thresholds)

    # prepare variables for storing all precision-recall values
    general_measures = [[] for _ in range(n_thresholds)]
    class_measures = [[] for _ in range(n_thresholds)]

    # compute precision-recall measures for each min_prediction_score threshold
    for thres in range(n_thresholds):

        # Counters for computing general precision-recall curve
        FP = 0
        TP = 0
        FN = 0
        total_GT = 0
        total_pred = 0

        # Counters for computing class-specific precision-recall curve
        TP_classes = np.zeros((n_classes,))
        FP_classes = np.zeros((n_classes,))
        FN_classes = np.zeros((n_classes,))
        total_GT_classes = np.zeros((n_classes,))
        total_pred_classes = np.zeros((n_classes,))

        for s in range(n_samples):

            # Re-use report of the current image provided in the parameters (or recently computed)
            [predicted_bboxes, predicted_Y, predicted_scores] = pred_list[s]
            [GT_bboxes, GT_Y] = gts[s]

            # Filter bounding boxes which are above the current threshold 'thres'
            aux_predicted_bboxes = []
            aux_predicted_Y = []
            aux_predicted_scores = []
            for pos, score in list(enumerate(predicted_scores)):
                if score > thresholds[thres]:
                    aux_predicted_bboxes.append(predicted_bboxes[pos])
                    aux_predicted_Y.append(predicted_Y[pos])
                    aux_predicted_scores.append(predicted_scores[pos])
            predicted_bboxes = aux_predicted_bboxes
            predicted_Y = aux_predicted_Y
            predicted_scores = aux_predicted_scores

            # Compute TPs, FPs and FNs
            [TP_, FP_, FN_, TP_c, FP_c, FN_c] = _computeMeasures(fixedIoU, n_classes, predicted_bboxes, predicted_Y,
                                                                 predicted_scores, GT_bboxes, GT_Y)

            total_GT += len(GT_bboxes)
            total_pred += len(predicted_bboxes)
            TP += TP_
            FP += FP_
            FN += FN_

            for pos in range(len(TP_c)):
                TP_classes[pos] += TP_c[pos]
                FP_classes[pos] += FP_c[pos]
                FN_classes[pos] += FN_c[pos]

            for y in GT_Y:
                total_GT_classes[y] += 1
            for y in predicted_Y:
                total_pred_classes[y] += 1

        # Compute general precision / recall / accuracy measures
        if (TP + FP) == 0:
            precision = 0.0
        else:
            precision = float(TP) / (TP + FP)
        if (TP + FN) == 0:
            recall = 0.0
        else:
            recall = float(TP) / (TP + FN)
        if (FP + FN + TP) == 0:
            accuracy = 0.0
        else:
            accuracy = float(TP) / (FP + FN + TP)

        # Compute class-specific precision - recall
        precision_classes = np.zeros((n_classes,))
        recall_classes = np.zeros((n_classes,), dtype=int)
        accuracy_classes = np.zeros((n_classes,), dtype=int)
        for c in range(n_classes):
            TP = TP_classes[c]
            FP = FP_classes[c]
            FN = FN_classes[c]
            if (TP + FP) == 0:
                precision_classes[c] = 0.0
            else:
                precision_classes[c] = float(TP) / (TP + FP)
            if (TP + FN) == 0:
                recall_classes[c] = 0.0
            else:
                recall_classes[c] = float(TP) / (TP + FN)
            if (FP + FN + TP) == 0:
                accuracy_classes[c] = 0.0
            else:
                accuracy_classes[c] = float(TP) / (FP + FN + TP)

        # store results
        general_measures[thres] = [precision, recall, accuracy, total_GT, total_pred]
        class_measures[thres] = [precision_classes, recall_classes, accuracy_classes, total_GT_classes,
                                 total_pred_classes]

    # Compute average precision (AP) (adapted from PASCAL VOC evaluation code VOCap.m)
    prec = [general_measures[thres][0] for thres in range(n_thresholds)][::-1]
    rec = [general_measures[thres][1] for thres in range(n_thresholds)][::-1]
    AP = _computeAP(prec, rec)

    # for thres in range(n_thresholds):
    #     logger.info(
    #         'Evaluation results (score >= %0.2f):\n\tPrecision: %f\n\tRecall: %f\n
    #         \tAccuracy: %f\n\tSamples GT: %d\n\tSamples predicted: %d' %
    #         (thresholds[thres],
    #          general_measures[thres][0], general_measures[thres][1], general_measures[thres][2],
    #          general_measures[thres][3], general_measures[thres][4]))

    if verbose > 0:
        logger.info('Average Precision (AP): %f' % AP)
        logger.info(
            'Evaluation results (score >= %0.2f):\n\tPrecision: %f\n\tRecall: %f\n'
            '\tAccuracy: %f\n\tSamples GT: %d\n\tSamples predicted: %d' %
            (thresholds[5],
             general_measures[5][0], general_measures[5][1], general_measures[5][2],
             general_measures[5][3], general_measures[5][4]))

    return_dict = dict()
    return_dict['AP'] = AP
    for thres in range(n_thresholds):
        return_dict['precision_' + str(thres)] = general_measures[thres][0]
        return_dict['recall_' + str(thres)] = general_measures[thres][1]
        return_dict['accuracy_' + str(thres)] = general_measures[thres][2]
        return_dict['total_GT_' + str(thres)] = general_measures[thres][3]
        return_dict['total_pred_' + str(thres)] = general_measures[thres][4]
        for c in range(n_classes):
            return_dict['precision_' + str(thres) + '_' + str(c)] = class_measures[thres][0][c]
            return_dict['recall_' + str(thres) + '_' + str(c)] = class_measures[thres][1][c]
            return_dict['accuracy_' + str(thres) + '_' + str(c)] = class_measures[thres][2][c]
            return_dict['total_GT_' + str(thres) + '_' + str(c)] = class_measures[thres][3][c]
            return_dict['total_pred_' + str(thres) + '_' + str(c)] = class_measures[thres][4][c]

    return return_dict


def _computeAP(prec, rec):
    AP = 0.0
    n_thresholds = len(prec)
    prec = [0] + prec + [0]
    rec = [0] + rec + [1]
    for i in range(n_thresholds, -1, -1):
        prec[i] = max(prec[i], prec[i + 1])
    for i in range(n_thresholds + 1):
        if rec[i + 1] != rec[i]:
            # increment of recall times precision
            AP += (rec[i + 1] - rec[i]) * prec[i]
    return AP


def _computeMeasures(IoU, n_classes, predicted_bboxes, predicted_Y, predicted_scores, GT_bboxes, GT_Y):
    """
    Computes TP, FP, and FN given a set of GT and Prediction BBoxes

    :param IoU:
    :param n_classes:
    :param predicted_bboxes:
    :param predicted_Y:
    :param predicted_scores:
    :param GT_bboxes:
    :param GT_Y:
    :return:
    """
    # General counters (without applying class distinctions)
    TP = 0
    FP = 0
    FN = 0

    # Class-specific counters
    TP_classes = np.zeros(n_classes)
    FP_classes = np.zeros(n_classes)
    FN_classes = np.zeros(n_classes)

    if len(predicted_Y) > 0:
        using_recognition = True
    else:
        using_recognition = False

    # Compute IoU for each pair of bounding boxes in (GT, pred)
    iou_values = []
    pred_ids = []
    match_bboxes = []
    for i, gt in list(enumerate(GT_bboxes)):
        for j, pred in list(enumerate(predicted_bboxes)):
            # compute IoU
            iou_values.append(computeIoU(gt, pred))
            pred_ids.append(j)
            match_bboxes.append([i, j])

    # Compute matchings (sorted by IoU)
    final_matches = []  # stores the final indices for [gt,pred] matches
    matched_gt = [False for _ in range(len(GT_bboxes))]
    matched_pred = [False for _ in range(len(predicted_bboxes))]
    # max_iou = np.argsort(np.array(iou_values, dtype=np.float))[::-1]
    max_scores = np.argsort(np.array(predicted_scores, dtype=np.float))[::-1]

    # Sort predictions by "scores"
    i = 0
    while i < len(max_scores) and not all(matched_gt):
        # m = match_bboxes[max_iou[i]]
        this_pred_id = max_scores[i]
        m_list = [[p_, match_bboxes[p_]] for p_, p in list(enumerate(pred_ids)) if p == this_pred_id]
        this_iou = [iou_values[p] for p, m in m_list]
        max_iou = np.argsort(np.array(this_iou, dtype=np.float))[::-1]

        # Sort GT by IoU
        j = 0
        while j < len(max_iou) and not matched_pred[this_pred_id]:  # if pred has not been matched yet
            j_ind = max_iou[j]
            if this_iou[j_ind] > IoU and not matched_gt[m_list[j_ind][1][0]]:
                # Assign match
                matched_gt[m_list[j_ind][1][0]] = True
                matched_pred[this_pred_id] = True
                final_matches.append(m_list[j_ind][1])
            j += 1
        i += 1

    # Compute FPs, FNs and TPs on the current image
    for m in matched_gt:
        if m:
            TP += 1
        else:
            FN += 1
    for m in matched_pred:
        if not m:
            FP += 1

    # Compute class-specific counters
    if using_recognition:
        # Check matching pairs
        for m in final_matches:
            y_gt = GT_Y[m[0]]
            y_pred = predicted_Y[m[1]]

            # GT and pred coincide
            if y_gt == y_pred:
                TP_classes[y_gt] += 1
            # Missclassified but correctly localized
            else:
                FN_classes[y_gt] += 1
                FP_classes[y_pred] += 1
        # Check missed GT bboxes
        for i, m in list(enumerate(matched_gt)):
            if not m:
                FN_classes[GT_Y[i]] += 1
        # Check mislocalized Pred bboxes
        for i, m in list(enumerate(matched_pred)):
            if not m:
                FP_classes[predicted_Y[i]] += 1

    return [TP, FP, FN, TP_classes, FP_classes, FN_classes]


def compute_perplexity(y_pred, y_true, verbose, split, mask=None):
    """
    Computes perplexity
    :param y_pred:
    :param y_true:
    :param verbose:
    :param split:
    :param mask:
    :return:
    """

    if mask is not None:
        y_pred /= np.sum(y_pred, axis=-1, keepdims=True)
        mask = np.reshape(np.asarray(mask), np.asarray(y_true).shape[:-1])[:, :, None]
        truth_mask = (y_true * mask).flatten().nonzero()[0]
        predictions = y_pred.flatten()[truth_mask]
        ppl = np.power(2, np.mean(-np.log2(predictions)))
        if verbose > 0:
            logger.info('Computing perplexity scores on the %s split...' % split)
            logger.info('PPL: ' + str(ppl))
        return ppl
    else:
        ppl = np.power(2, np.mean(-np.log2(y_pred)))
        if verbose > 0:
            logger.info('Computing perplexity scores on the %s split...' % split)
            logger.info('PPL: ' + str(ppl))
        return ppl


# AUXILIARY FUNCTIONS

def vqa_store(question_id_list, answer_list, path):
    """
    Saves the answers on question_id_list in the VQA-like format.

    In:
        question_id_list - list of the question ids
        answer_list - list with the answers
        path - path where the file is saved
    """
    question_answer_pairs = []
    if len(question_id_list) != len(answer_list):
        raise AssertionError('must be the same number of questions and answers')
    for q, a in list(zip(question_id_list, answer_list)):
        question_answer_pairs.append({'question_id': q, 'answer': str(a)})
    with open(path, 'w') as f:
        json.dump(question_answer_pairs, f)


def caption_store(samples, path):
    """
    Save the samples into a file.
    :param samples: Samples generated.
    :param path: Path to the output file.
    :return:
    """
    with open(path, 'w') as f:
        print('\n'.join(samples), file=f)


# List of evaluation functions and their identifiers (will be used in params['METRICS'])
selectMetric = {
    'vqa': eval_vqa,  # Metric for the VQA challenge
    'coco': get_coco_score,  # MS COCO evaluation library (BLEU, METEOR and CIDEr scores)
    'multilabel_metrics': multilabel_metrics,  # Set of multilabel classification metrics from sklearn
    'multiclass_metrics': multiclass_metrics,  # Set of multiclass classification metrics from sklearn
    'AP': averagePrecision,
    'sem_seg_acc': semantic_segmentation_accuracy,
    'sem_seg_iou': semantic_segmentation_meaniou,
    'ppl': compute_perplexity,
}

select = selectMetric
